#include "t3f/t3f.h"
#include "modules/input.h"
#include "gameplay/game_intro.h"
#include "instance.h"

/* user pressed close button (or ESC key in certain places) */
void wl_close(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->state != WL_STATE_GLOBAL_PAUSE && instance->exit_tick <= 0)
	{
		if(instance->state == WL_STATE_GAME_INTRO || instance->state == WL_STATE_GAME_WIN || instance->state == WL_STATE_MUSIC)
		{
			t3f_pause_music();
			wl_set_music_fade_speed(instance->music_queue, 0.0);
		}
		instance->global_pause_state = instance->state;
		t3f_clear_key_states();
		t3f_clear_mouse_state();
		t3f_clear_touch_state();
		wl_reset_particle_list(instance->particle_list[0]);
		instance->menu[instance->current_menu]->hover_element = -1;
		instance->menu[WL_MENU_GAME_EXIT]->hover_element = -1;
		instance->state = WL_STATE_GLOBAL_PAUSE;
	}
}

static void _wl_handle_joystick_event(ALLEGRO_EVENT * ep, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(ep->type == ALLEGRO_EVENT_JOYSTICK_AXIS)
	{
		if(fabs(ep->joystick.pos) >= instance->dead_zone)
		{
			wl_select_input(WL_INPUT_TYPE_GAMEPAD, t3f_get_joystick_number(ep->joystick.id), data);
		}
	}
	else if(ep->type == ALLEGRO_EVENT_JOYSTICK_BUTTON_DOWN)
	{
		wl_select_input(WL_INPUT_TYPE_GAMEPAD, t3f_get_joystick_number(ep->joystick.id), data);
	}
	else if(ep->type == ALLEGRO_EVENT_JOYSTICK_CONFIGURATION)
	{
		int i;

		if(al_get_num_joysticks() > 0 || instance->input_device_handle)
		{
			for(i = 0; i < al_get_num_joysticks(); i++)
			{
				if(instance->input_device_handle == al_get_joystick(i))
				{
					break;
				}
			}
			if(i == al_get_num_joysticks())
			{
				instance->input_device = -1;
				instance->input_device_handle = NULL;
				instance->input_device_lost = true;
			}
		}
	}
}

void wl_process_event(ALLEGRO_EVENT * event, void * data)
{
	switch(event->type)
	{
		case ALLEGRO_EVENT_JOYSTICK_AXIS:
		case ALLEGRO_EVENT_JOYSTICK_BUTTON_DOWN:
		case ALLEGRO_EVENT_JOYSTICK_BUTTON_UP:
		{
			_wl_handle_joystick_event(event, data);
			t3f_event_handler(event);
			break;
		}
		case ALLEGRO_EVENT_MOUSE_AXES:
		case ALLEGRO_EVENT_MOUSE_BUTTON_DOWN:
		case ALLEGRO_EVENT_MOUSE_BUTTON_UP:
		{
			wl_select_input(WL_INPUT_TYPE_MOUSE, 0, data);
			t3f_event_handler(event);
			break;
		}
		case ALLEGRO_EVENT_TOUCH_BEGIN:
		case ALLEGRO_EVENT_TOUCH_END:
		case ALLEGRO_EVENT_TOUCH_MOVE:
		case ALLEGRO_EVENT_TOUCH_CANCEL:
		{
			wl_select_input(WL_INPUT_TYPE_TOUCH, 0, data);
			t3f_event_handler(event);
			break;
		}
		case ALLEGRO_EVENT_KEY_DOWN:
		case ALLEGRO_EVENT_KEY_UP:
		{
			wl_select_input(WL_INPUT_TYPE_KEYBOARD, 0, data);
			t3f_event_handler(event);
			break;
		}
		case ALLEGRO_EVENT_DISPLAY_CLOSE:
		{
			wl_close(data);
			break;
		}

		/* pass the event through to T3F for handling by default */
		default:
		{
			t3f_event_handler(event);
			break;
		}
	}
}
