#include "t3f/t3f.h"
#include "t3f/resource.h"
#include "modules/data.h"
#include "modules/text_entry.h"
#include "modules/leaderboards.h"
#include "modules/level_list.h"
#include "instance.h"
#include "preferences.h"

void wl_free_data(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;

	if(instance->cinema)
	{
		wl_destroy_cinema(instance->cinema);
		instance->cinema = NULL;
	}
	for(i = 0; i < WL_MAX_ANIMATIONS; i++)
	{
		if(instance->animation[i])
		{
			t3f_destroy_animation(instance->animation[i]);
			instance->animation[i] = NULL;
		}
	}
	if(instance->box_nine_patch)
	{
		np_destroy_nine_patch_bitmap(instance->box_nine_patch);
	}
	for(i = 0; i < WL_MAX_BITMAPS; i++)
	{
		if(instance->bitmap[i])
		{
			t3f_destroy_bitmap(instance->bitmap[i]);
			instance->bitmap[i] = NULL;
		}
	}
	for(i = 0; i < WL_MAX_SAMPLES; i++)
	{
		if(instance->sample[i])
		{
			al_destroy_sample(instance->sample[i]);
			instance->sample[i] = NULL;
		}
	}
	for(i = 0; i < WL_MAX_FONTS; i++)
	{
		if(instance->font[i])
		{
			t3f_destroy_resource(instance->font[i]);
			instance->font[i] = NULL;
		}
	}
	for(i = 0; i < WL_MAX_ATLASES; i++)
	{
		if(instance->atlas[i])
		{
			t3f_destroy_atlas(instance->atlas[i]);
			instance->atlas[i] = NULL;
		}
	}
	wl_destroy_dictionary(instance->dictionary);
	instance->dictionary = NULL;
	wl_destroy_profile(instance->profile);
	instance->profile = NULL;
	wl_destroy_effects_data(instance->effects_data);
	instance->effects_data = NULL;
	wl_destroy_music_queue(instance->music_queue);
	instance->music_queue = NULL;
	wl_destroy_particle_list(instance->particle_list[0]);
	instance->particle_list[0] = NULL;
	wl_destroy_particle_list(instance->particle_list[1]);
	instance->particle_list[1] = NULL;
	wl_destroy_shape_lists(instance->shape_lists);
	instance->shape_lists = NULL;
	wl_destroy_text_entry_data(instance->text_entry);
	instance->text_entry = NULL;
	wl_destroy_level_list(instance->level_list);
	instance->level_list = NULL;
	wl_destroy_playlist(instance->playlist);
	instance->playlist = NULL;
	t3f_destroy_object_loader(instance->leaderboard_loader);
	instance->leaderboard_loader = NULL;
}

void wl_exit(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char fnbuf[1024];

	t3f_save_achievements_data(instance->achievements, t3f_user_data, "Achievements");
	wl_save_preferences(data);
	wl_save_profile(instance->profile, t3f_get_filename(t3f_data_path, "profile.dat", fnbuf, 1024));
	wl_title_free_data(data);
	wl_free_data(data);
	wl_title_exit(data);
	wl_free_leaderboard_data(data);
}
