#include "t3f/rng.h"
#include "modules/data.h"
#include "modules/effects.h"
#include "instance.h"
#include "game.h"
#include "game_sprite.h"
#include "game_hazard.h"

void wl_setup_board_tiles(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j;

	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			if(i == 0 && j == 0)
			{
				instance->game.board[i][j].tile_index = WL_TILE_TOP_LEFT;
			}
			else if(i == 0 && j == WL_MAX_BOARD_WIDTH - 1)
			{
				instance->game.board[i][j].tile_index = WL_TILE_TOP_RIGHT;
			}
			else if(i == WL_MAX_BOARD_HEIGHT - 1 && j == 0)
			{
				instance->game.board[i][j].tile_index = WL_TILE_BOTTOM_LEFT;
			}
			else if(i == WL_MAX_BOARD_HEIGHT - 1 && j == WL_MAX_BOARD_WIDTH - 1)
			{
				instance->game.board[i][j].tile_index = WL_TILE_BOTTOM_RIGHT;
			}
			else if(i == 0)
			{
				instance->game.board[i][j].tile_index = WL_TILE_TOP;
			}
			else if(i == WL_MAX_BOARD_HEIGHT - 1)
			{
				instance->game.board[i][j].tile_index = WL_TILE_BOTTOM;
			}
			else if(j == 0)
			{
				instance->game.board[i][j].tile_index = WL_TILE_LEFT;
			}
			else if(j == WL_MAX_BOARD_WIDTH - 1)
			{
				instance->game.board[i][j].tile_index = WL_TILE_RIGHT;
			}
			else if(i == 0)
			{
				instance->game.board[i][j].tile_index = WL_TILE_TOP;
			}
			else if(i == WL_MAX_BOARD_HEIGHT - 1)
			{
				instance->game.board[i][j].tile_index = WL_TILE_BOTTOM;
			}
			else
			{
				instance->game.board[i][j].tile_index = WL_TILE_CENTER;
			}
		}
	}

}

void wl_clear_board_flags(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j;
	
	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			instance->game.board[i][j].flags = 0;
			instance->game.board[i][j].used = false;
		}
	}
}

void wl_board_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j, type;
	ALLEGRO_COLOR col_base[3];
	ALLEGRO_COLOR col_outrange[3];
	ALLEGRO_COLOR col_used[3];
	
	col_base[0] = al_map_rgba(162, 180, 213, 204);
	col_outrange[0] = wl_darken_color(col_base[0], 0.6);
	col_used[0] = wl_darken_color(col_base[0], 0.4);
	col_base[1] = al_map_rgba(255, 223, 96, 204);
	col_outrange[1] = wl_darken_color(col_base[1], 0.6);
	col_used[1] = wl_darken_color(col_base[1], 0.4);
	col_base[2] = al_map_rgba(255, 155, 155, 204);
	col_outrange[2] = wl_darken_color(col_base[2], 0.6);
	col_used[2] = wl_darken_color(col_base[2], 0.4);
	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			
			if(instance->game.board[i][j].flags & WL_BOARD_FLAG_HAZARD)
			{
				type = 2;
			}
			else if(instance->game.board[i][j].flags & WL_BOARD_FLAG_POWERUP)
			{
				type = 1;
			}
			else
			{
				type = 0;
			}
			/* blank tile where player is standing */
			if(instance->game.player[0].x == j && instance->game.player[0].y == i && instance->game.player[0].state == WL_PLAYER_STATE_NORMAL)
			{
				instance->game.board[i][j].type = 3;
				instance->game.board[i][j].color = col_used[type];
			}
			
			/* blank tile for spaces with no letters */
			else if(instance->game.board[i][j].letter == ' ')
			{
				instance->game.board[i][j].type = 3;
				instance->game.board[i][j].color = col_used[type];
			}
			
			/* in range */
			else if(abs(instance->game.player[0].x - j) < instance->game.player[0].range && abs(instance->game.player[0].y - i) < instance->game.player[0].range)
			{
				instance->game.board[i][j].type = 2;
				instance->game.board[i][j].color = col_base[type];
			}
			
			/* out of range */
			else
			{
				instance->game.board[i][j].type = 0;
				instance->game.board[i][j].color = col_outrange[type];
			}
		}
	}
}

static void _wl_board_render_normal(WL_INSTANCE * instance)
{
	int i, j;
	char buf[4] = {0};

	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			wl_draw_single_tile(instance->tile[WL_TILE_BOARD], instance->game.board[i][j].tile_index, instance->game.board[i][j].color, WL_BOARD_X + j * instance->tile[WL_TILE_BOARD]->width, WL_BOARD_Y + i * instance->tile[WL_TILE_BOARD]->height, 0.0, 0);
		}
	}
	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			switch(instance->game.board[i][j].letter)
			{
				/* don't render character for powerup tiles */
				case '?':
				case '!':
				case '_':
				case '0':
				{
					buf[0] = '\0';
					break;
				}
				case 'Q':
				{
					buf[0] = 'Q';
					buf[1] = 'u';
					break;
				}
				default:
				{
					buf[0] = instance->game.board[i][j].letter;
					buf[1] = '\0';
					break;
				}
			}
			t3f_draw_scaled_text(instance->font[WL_FONT_32], al_map_rgba_f(0.0, 0.0, 0.0, 1.0), WL_BOARD_X + j * WL_BOARD_TILE_SIZE + WL_BOARD_TILE_SIZE / 2, WL_BOARD_Y + i * WL_BOARD_TILE_SIZE + 2.0, 0, 1.0, T3F_FONT_ALIGN_CENTER, buf);
		}
	}
}

void wl_board_render(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_wl_board_render_normal(instance);
	
}

static char wl_board_retrieve_powerup(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, r;
	char c = 0;
	
	if(instance->game.powerups)
	{
		r = rand() % instance->game.powerups;
		switch(instance->game.powerup[r])
		{
			case WL_POWERUP_GURU:
			{
				c = '?';
				break;
			}
			case WL_POWERUP_HOURGLASS:
			{
				c = '_';
				break;
			}
			case WL_POWERUP_LIFE:
			{
				c = '!';
				break;
			}
		}
		for(i = r; i < instance->game.powerups - 1; i++)
		{
			instance->game.powerup[i] = instance->game.powerup[i + 1];
		}
		instance->game.powerups--;
		if(instance->game.powerups <= 0)
		{
			wl_game_generate_powerups(&instance->game); // generate new list of powerups when we run out
		}
	}
	return c;
}

char wl_board_generate_letter(bool allow_powerups, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float r, a;
	int i;
	
	/* see if we need to generate a powerup */
	if(instance->game.uses_powerups && allow_powerups)
	{
		instance->game.powerup_count--;
		if(instance->game.powerup_count <= 0)
		{
			/* set amount of "time" it takes to generate a powerup */
			instance->game.powerup_count = WL_POWERUP_COUNT;
			
			/* destroy other powerups on the board, we only want one at a time */
			for(i = 0; i < WL_MAX_SPRITES; i++)
			{
				if(instance->game.sprite[i].active)
				{
					if(instance->game.sprite[i].type == WL_SPRITE_TYPE_POWERUP_GURU || instance->game.sprite[i].type == WL_SPRITE_TYPE_POWERUP_HEALTH || instance->game.sprite[i].type == WL_SPRITE_TYPE_POWERUP_SNEAKER)
					{
						wl_explode_sprite(&instance->game.sprite[i], data);
						instance->game.board[(int)instance->game.sprite[i].y / WL_BOARD_TILE_SIZE][(int)instance->game.sprite[i].x / WL_BOARD_TILE_SIZE].letter = ' ';
						t3f_play_sample(instance->sample[WL_SAMPLE_DESTROY], 1.0, 0.0, 1.0);
						break;
					}
				}
			}
			
			return wl_board_retrieve_powerup(data);
		}
	}
	
	/* generate a letter */
	a = 0.0;
	for(i = 0; i < 26; i++)
	{
		a += wl_letter_probability[i];
	}
	r = t3f_drand(&instance->game.rng_state) * a;
	a = 0.0;
	for(i = 0; i < 26; i++)
	{
		a += wl_letter_probability[i];
		if(r < a)
		{
			return 'A' + i;
		}
	}
	return 'Z';
}

void wl_board_generate_powerup_sprite(int i, int j, int letter, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int s;
	
	switch(letter)
	{
		case '!':
		{
			s = wl_game_add_sprite(&instance->game, WL_SPRITE_TYPE_POWERUP_HEALTH);
			if(s >= 0)
			{
				instance->game.sprite[s].x = j * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].y = i * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].tick = 0;
				instance->game.sprite[s].timer = i * WL_MAX_BOARD_WIDTH + j;
			}
			break;
		}
		case '?':
		{
			s = wl_game_add_sprite(&instance->game, WL_SPRITE_TYPE_POWERUP_GURU);
			if(s >= 0)
			{
				instance->game.sprite[s].x = j * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].y = i * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].timer = i * WL_MAX_BOARD_WIDTH + j;
			}
			break;
		}
		case '_':
		{
			s = wl_game_add_sprite(&instance->game, WL_SPRITE_TYPE_POWERUP_SNEAKER);
			if(s >= 0)
			{
				instance->game.sprite[s].x = j * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].y = i * WL_BOARD_TILE_SIZE;
				instance->game.sprite[s].a = 0.0;
				instance->game.sprite[s].va = 0.0;
				instance->game.sprite[s].timer = i * WL_MAX_BOARD_WIDTH + j;
			}
			break;
		}
	}
}

static bool wl_is_player_space(int x, int y, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(x != instance->game.player[0].x || y != instance->game.player[0].y)
	{
		return false;
	}
	return true;
}

void wl_board_generate_new_letters(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j;
	
	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			if(instance->game.board[i][j].letter == ' ' && !wl_is_player_space(j, i, data))
			{
				instance->game.board[i][j].letter = wl_board_generate_letter(true, data);
				wl_board_generate_powerup_sprite(i, j, instance->game.board[i][j].letter, data);
			}
		}
	}
}

/* used when continuing the game, generates new tiles for all non-letter tiles */
void wl_board_regenerate(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j;
	
	for(i = 0; i < WL_MAX_BOARD_HEIGHT; i++)
	{
		for(j = 0; j < WL_MAX_BOARD_WIDTH; j++)
		{
			switch(instance->game.board[i][j].letter)
			{
				case ' ':
				case '!':
				case '?':
				case '2':
				case '3':
				case '0':
				{
					instance->game.board[i][j].letter = wl_board_generate_letter(false, data);
					break;
				}
			}
		}
	}
}

void wl_add_board_particle(ALLEGRO_COLOR color, int x, int y, int life, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	wl_add_particle(instance->particle_list[1], WL_PARTICLE_TYPE_FLASH, instance->animation[WL_ANIMATION_PARTICLE], color, WL_BOARD_X + x * WL_BOARD_TILE_SIZE, WL_BOARD_Y + y * WL_BOARD_TILE_SIZE, 0.0, WL_BOARD_TILE_SIZE, life, 0);
}

void wl_generate_board_particles(ALLEGRO_COLOR color, int x, int y, int r, void * data)
{
	int i, j, sx, sy, ex, ey;
	
	for(i = 1; i < r; i++)
	{
		sx = x - i;
		sy = y - i;
		ex = x + i;
		ey = y + i;
		for(j = sx; j <= ex; j++)
		{
			if(sy >= 0 && j >= 0 && j < WL_MAX_BOARD_WIDTH)
			{
				wl_add_board_particle(color, j, sy, 10 + i * 2, data);
			}
			if(ey < WL_MAX_BOARD_HEIGHT && j >= 0 && j < WL_MAX_BOARD_WIDTH)
			{
				wl_add_board_particle(color, j, ey, 10 + i * 2, data);
			}
		}
		for(j = sy + 1; j <= ey - 1; j++)
		{
			if(sx >= 0 && j >= 0 && j < WL_MAX_BOARD_HEIGHT)
			{
				wl_add_board_particle(color, sx, j, 10 + i * 2, data);
			}
			if(ex < WL_MAX_BOARD_WIDTH && j >= 0 && j < WL_MAX_BOARD_HEIGHT)
			{
				wl_add_board_particle(color, ex, j, 10 + i * 2, data);
			}
		}
	}
}
