#ifndef WL_GAME_H
#define WL_GAME_H

#include "t3f/rng.h"
#include "modules/particle.h"
#include "modules/dictionary.h"
#include "player.h"
#include "sprite.h"
#include "board.h"

#define WL_GAME_STATE_PLAY        0
#define WL_GAME_STATE_PLAYER_WAIT 1
#define WL_GAME_STATE_DONE        2

#define WL_LIFE_BG_COLOR          al_map_rgba(255, 50, 95, 255)
#define WL_LIFE_FILL_COLOR        al_map_rgba(0, 189, 82, 255)
#define WL_LIFE_FREEZE_BG_COLOR   al_map_rgba(101, 188, 255, 255)
#define WL_LIFE_FREEZE_FILL_COLOR al_map_rgba(193, 223, 255, 255)

#define WL_LIFE_DECREASE_GURU_MODE 3.6

#define WL_GAME_INIT_FLAG_NO_MUSIC            (1 << 0)
#define WL_GAME_INIT_FLAG_NO_LEVEL_LOAD       (1 << 1)
#define WL_GAME_INIT_FLAG_NO_BOARD_GENERATION (1 << 2)

typedef struct
{
	
	int mode;
	WL_BOARD_TILE board[WL_MAX_BOARD_HEIGHT][WL_MAX_BOARD_WIDTH];
	WL_PLAYER player[WL_MAX_PLAYERS];
	int powerup[WL_MAX_POWERUPS];
	int powerups;
	int powerup_count;
	int hazard_count;
	WL_SPRITE sprite[WL_MAX_SPRITES];
	WL_SPRITE_LIST sprite_list;
	bool finished, retry, regenerate;
	int base_range;
	int powerup_health;
	int powerup_guru;
	int powerup_sneaker;

	/* guru stuff */
	bool guru_active;
	float guru_z, guru_vz, guru_dir;
	float guru_x, guru_y;
	int guru_state;
	int guru_timer;
	ALLEGRO_COLOR guru_color;
	
	/* game mode feature toggles */
	bool uses_powerups;
	bool uses_hazards;
	bool uses_life;
	bool uses_level_progress;
	bool uses_invalid;
	
	int time;
	int tick;
	bool got_new_level;
	double music_pos, old_music_pos;
	double music_length;
	bool fast_forward;
	float music_fade_level;
	bool music_done;
	float word_list_ox;
	int want_new_level;
	int current_best_word;
	int current_unused_letters;
	int wipe_count; // >= 0 means we are processing a wipe
	T3F_RNG_STATE rng_state;
	int players;
	int current_player;
	int state;
	
} WL_GAME;

void wl_game_generate_powerups(WL_GAME * game);
int wl_get_best_word(void * data);
int wl_game_figure_base_points(const char * word);
int wl_game_figure_total_points(WL_DICTIONARY * dictionary, int w);
void wl_generate_board_particles(ALLEGRO_COLOR color, int x, int y, int r, void * data);

void wl_game_setup_type(void * data);
bool wl_game_initialize(void * data, int flags);
void wl_game_scroll_level_name(void * data);
void wl_game_exit(void * data);
void wl_game_end(void * data);
void wl_game_update_best_flags(void * data);
void wl_game_process_bg(void * data);
void wl_game_guru_logic(void * data);
void wl_game_process_overlay_menu(T3F_GUI * mp, void * data);
void wl_game_logic(void * data);
void wl_game_render_bg(void * data);
void wl_game_render(void * data);
void wl_game_paused_logic(void * data);
void wl_game_paused_render(void * data);
void wl_game_credits_logic(void);
void wl_game_credits_render(void);
void wl_game_win_logic(void * data);
void wl_game_win_render(void * data);
void wl_demo_win_logic(void * data);
void wl_demo_win_render(void * data);

#endif
