#include "t3f/t3f.h"
#include "modules/level.h"
#include "instance.h"

bool wl_level_transition_complete(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	ALLEGRO_BITMAP * bp = NULL;
	int i;

	if(!instance->level[0]->shapes_used)
	{
		return false;
	}
	for(i = 0; i < WL_MAX_SHAPES; i++)
	{
		if(instance->shape_lists->bg_shape[i].flags)
		{
			if(bp == NULL)
			{
				bp = *instance->shape_lists->bg_shape[i].parent_bp;
			}

			/* if we come across a different bitmap, switch lists */
			if(*instance->shape_lists->bg_shape[i].parent_bp != bp)
			{
				return false;
			}
		}
	}
	bp = NULL;
	for(i = 0; i < WL_MAX_SHAPES; i++)
	{
		if(instance->shape_lists->fg_shape[i].flags)
		{
			if(bp == NULL)
			{
				bp = *instance->shape_lists->fg_shape[i].parent_bp;
			}

			/* if we come across a different bitmap, switch lists */
			if(*instance->shape_lists->fg_shape[i].parent_bp != bp)
			{
				return false;
			}
		}
	}
	return true;
}

void wl_game_level_bg_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	bool generate;
	int ps;
	int i;

	wl_update_level(instance->level[0]);
	if(instance->level[1])
	{
		wl_update_level(instance->level[1]);
	}
	generate = true;
	for(i = 0; i < WL_MAX_SHAPES; i++)
	{
		if(instance->shape_lists->bg_shape[i].flags)
		{
			wl_process_shape(&instance->shape_lists->bg_shape[i]);
			if(instance->shape_lists->bg_shape[i].ox <= -64.0)
			{
				generate = false;
			}
		}
	}
	if(generate)
	{
		ps = instance->current_bg_shape - 1;
		if(ps < 0)
		{
			ps = WL_MAX_SHAPES - 1;
		}
		wl_connect_shape(&instance->shape_lists->bg_shape[instance->current_bg_shape], &instance->shape_lists->bg_shape[ps], &instance->level[0]->shape[rand() % instance->level[0]->shapes], &instance->level[0]->shape_image->bitmap);
		instance->current_bg_shape++;
		if(instance->current_bg_shape >= WL_MAX_SHAPES)
		{
			instance->current_bg_shape = 0;
		}
		instance->level[0]->shapes_used = true;
	}

	generate = true;
	for(i = 0; i < WL_MAX_SHAPES; i++)
	{
		if(instance->shape_lists->fg_shape[i].flags)
		{
			wl_process_shape(&instance->shape_lists->fg_shape[i]);
			if(instance->shape_lists->fg_shape[i].ox <= -32.0)
			{
				generate = false;
			}
		}
	}
	if(generate)
	{
		ps = instance->current_fg_shape - 1;
		if(ps < 0)
		{
			ps = WL_MAX_SHAPES - 1;
		}
		wl_connect_shape(&instance->shape_lists->fg_shape[instance->current_fg_shape], &instance->shape_lists->fg_shape[ps], &instance->level[0]->shape[rand() % instance->level[0]->shapes], &instance->level[0]->shape_image->bitmap);
		instance->current_fg_shape++;
		if(instance->current_fg_shape >= WL_MAX_SHAPES)
		{
			instance->current_fg_shape = 0;
		}
		instance->level[0]->shapes_used = true;
	}

	/* process bg image */
	instance->level[0]->z -= instance->level[0]->vz;
	instance->level[0]->a += instance->level[0]->va;
	if(instance->level[1])
	{
		instance->level[1]->z -= instance->level[1]->vz;
		instance->level[1]->a += instance->level[1]->va;
	}
}

void wl_game_level_bg_render(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;
	ALLEGRO_COLOR color;
	ALLEGRO_BITMAP * bp = NULL;
	int list = 0;
	float a;

	if(instance->game.tick > 0)
	{
		a = (float)(instance->game.tick) / WL_LEVEL_FADE_FRAMES;
		color = wl_fade_color(instance->level[0]->color, instance->level[1]->color, a);
	}
	else
	{
		color = instance->level[0]->color;
	}


	/* draw background */
	if(instance->level[0]->bg_animation)
	{
		t3f_draw_rotated_animation(instance->level[0]->bg_animation, al_map_rgba_f(1.0, 1.0, 1.0, 1.0), 0, instance->level[0]->cx, instance->level[0]->cy, instance->level[0]->ox, instance->level[0]->oy, instance->level[0]->z, instance->level[0]->a, 0);
		if(instance->level[1] && instance->level[1]->bg_animation && instance->game.tick > 0)
		{
			t3f_draw_rotated_animation(instance->level[1]->bg_animation, al_map_rgba_f(a, a, a, a), 0, instance->level[1]->cx, instance->level[1]->cy, instance->level[1]->ox, instance->level[1]->oy, instance->level[1]->z, instance->level[1]->a, 0);
		}
	}

	wl_clear_cached_shapes(instance->shape_lists, 0);
	wl_clear_cached_shapes(instance->shape_lists, 1);
	for(i = 0; i < instance->shape_lists->shapes; i++)
	{
		if(instance->shape_lists->bg_shape[i].flags)
		{
			if(bp == NULL)
			{
				bp = *instance->shape_lists->bg_shape[i].parent_bp;
			}

			/* if we come across a different bitmap, switch lists */
			if(*instance->shape_lists->bg_shape[i].parent_bp != bp)
			{
				bp = *instance->shape_lists->bg_shape[i].parent_bp;
				list = 1 - list;
			}
			wl_cache_shape(instance->shape_lists, list, &instance->shape_lists->bg_shape[i]);
		}
	}
	wl_draw_cached_shapes(instance->shape_lists, 0, wl_darken_color(color, 0.75));
	wl_draw_cached_shapes(instance->shape_lists, 1, wl_darken_color(color, 0.75));
	wl_clear_cached_shapes(instance->shape_lists, 0);
	wl_clear_cached_shapes(instance->shape_lists, 1);
	bp = NULL;
	list = 0;
	for(i = 0; i < WL_MAX_SHAPES; i++)
	{
		if(instance->shape_lists->fg_shape[i].flags)
		{
			if(bp == NULL)
			{
				bp = *instance->shape_lists->fg_shape[i].parent_bp;
			}

			/* if we come across a different bitmap, switch lists */
			if(*instance->shape_lists->fg_shape[i].parent_bp != bp)
			{
				bp = *instance->shape_lists->fg_shape[i].parent_bp;
				list = 1 - list;
			}
			wl_cache_shape(instance->shape_lists, list, &instance->shape_lists->fg_shape[i]);
		}
	}
	wl_draw_cached_shapes(instance->shape_lists, 0, color);
	wl_draw_cached_shapes(instance->shape_lists, 1, color);
}
