#include "t3f/t3f.h"
#include "modules/particle.h"
#include "modules/music_queue.h"
#include "modules/effects.h"
#include "modules/data.h"
#include "ui/menu_overlay.h"
#include "instance.h"
#include "game_level.h"

void wl_game_music_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;
	int f = 1;
	WL_LEVEL * lev;

	if(!instance->level[1] && instance->playlist_pos < instance->playlist->count - 1)
	{
		instance->level[1] = wl_load_level(instance->playlist->entry[instance->playlist_pos + 1], WL_LEVEL_LOAD_FLAG_THREADED, data);
	}

	wl_process_particle_list(instance->particle_list[1]);
	if(instance->game.fast_forward)
	{
		f = 8;
		if(t3f_get_music_state() == T3F_MUSIC_STATE_PLAYING)
		{
			al_set_audio_stream_speed(t3f_stream, 8.0);
		}
		wl_set_music_fade_speed(instance->music_queue, WL_MUSIC_FADE_SPEED * 8.0);
	}
	else
	{
		if(t3f_get_music_state() == T3F_MUSIC_STATE_PLAYING)
		{
			al_set_audio_stream_speed(t3f_stream, 1.0);
			t3f_set_music_volume(t3f_get_music_volume());
		}
		wl_set_music_fade_speed(instance->music_queue, WL_MUSIC_FADE_SPEED);
	}
	if(instance->game.fast_forward && !instance->input_state.action.held)
	{
		instance->game.fast_forward = false;
	}
	for(i = 0; i < f; i++)
	{
		wl_game_level_bg_logic(data);
	}

	/* add time to profile */
	instance->profile->time++;

	if(instance->input_state.pause.pressed)
	{
		wl_menu_proc_overlay_menu(data, 0, NULL);
	}
	if(t3f_get_music_state() == T3F_MUSIC_STATE_PLAYING)
	{

		/* get music length if it hasn't been gotten */
		if(instance->game.music_length <= 0.0)
		{
			instance->game.music_length = al_get_audio_stream_length_secs(t3f_stream);
		}

		instance->game.old_music_pos = instance->game.music_pos;
		instance->game.music_pos = al_get_audio_stream_position_secs(t3f_stream);
		if(instance->game.music_pos < 0.0)
		{
			instance->game.music_pos = 0.0;
		}

		/* if we have looped, go to next track */
		if(wl_get_music_queue_state(instance->music_queue) == WL_MUSIC_QUEUE_STATE_PLAYING && (instance->game.music_pos < instance->game.old_music_pos || instance->game.music_pos >= instance->game.music_length) && wl_level_transition_complete(data))
		{
			if(instance->playlist_pos < instance->playlist->count - 1)
			{
				instance->playlist_pos++;
				lev = instance->level[0];
				instance->level[0] = instance->level[1];
				instance->level[1] = lev;
				wl_play_level_music(instance->level[0], data);
				instance->game.tick = WL_LEVEL_FADE_FRAMES;
				instance->game.got_new_level = true;
			}

			/* fade out and go back to title screen */
			else
			{
				wl_finish_current_song(instance->music_queue, 0.0);
				instance->game.music_done = true;
			}
		}
	}
	else
	{
		instance->game.music_pos = 0.0;
		instance->game.old_music_pos = 0.0;
		instance->game.music_length = 0.0;
	}
	if(instance->state == WL_STATE_MUSIC && !instance->game.music_done)
	{
		wl_game_process_overlay_menu(instance->menu[WL_MENU_MUSIC_OVERLAY], data);
	}
	if(instance->game.tick > 0)
	{
		instance->game.tick -= f;
		if(instance->game.tick < 0)
		{
			instance->game.tick = 0;
		}
	}
	if(instance->game.tick <= 0 && instance->game.got_new_level && wl_level_transition_complete(data))
	{
		if(instance->level[1])
		{
			wl_destroy_level(instance->level[1]);
			instance->level[1] = NULL;
		}
		instance->game.got_new_level = false;
	}
	if(instance->game.music_done)
	{
		for(i = 0; i < f; i++)
		{
			instance->game.music_fade_level += 1.0 / (float)WL_LEVEL_FADE_FRAMES;
			if(instance->game.music_fade_level >= 1.0)
			{
				t3f_render(true);
				wl_game_exit(data);
				wl_play_music(instance->music_queue, "data/music/theme.ogg");
				wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), al_map_rgba_f(0.0, 0.0, 0.0, 0.0), 60);
				break;
			}
		}
	}
}

void wl_game_render_music_static_info(float offx, float offy, float offz, ALLEGRO_COLOR color, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float ox, oy;
	char buf[64] = {0};

	ox = t3f_default_view->left + 16 + t3f_get_text_width(instance->font[WL_FONT_ASCII], "Musician: ");
	oy = t3f_default_view->top + 12;

	sprintf(buf, "Title: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], color, ox + offx, oy + offy, offz, T3F_FONT_ALIGN_RIGHT, buf);
	oy += 24;

	sprintf(buf, "Artist: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], color, ox + offx, oy + offy, offz, T3F_FONT_ALIGN_RIGHT, buf);
	oy += 24;

	sprintf(buf, "Music: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], color, ox + offx, oy + offy, offz, T3F_FONT_ALIGN_RIGHT, buf);
	oy += 24;

	sprintf(buf, "Musician: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], color, ox + offx, oy + offy, offz, T3F_FONT_ALIGN_RIGHT, buf);
}

void wl_game_render_music_info(int level, float offx, float offy, float offz, ALLEGRO_COLOR color, float fade, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float ox, oy;
	char buf[64] = {0};

	ox = t3f_default_view->left + 16 + t3f_get_text_width(instance->font[WL_FONT_ASCII], "Musician: ");
	oy = t3f_default_view->top + 12;

	sprintf(buf, "Title: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], wl_alpha_color(color, fade), ox + offx, oy + offy, offz, 0, instance->level[level]->name);
	oy += 24;

	sprintf(buf, "Artist: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], wl_alpha_color(color, fade), ox + offx, oy + offy, offz, 0, instance->level[level]->artist);
	oy += 24;

	sprintf(buf, "Music: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], wl_alpha_color(color, fade), ox + offx, oy + offy, offz, 0, instance->level[level]->music_title);
	oy += 24;

	sprintf(buf, "Musician: ");
	t3f_draw_text(instance->font[WL_FONT_ASCII], wl_alpha_color(color, fade), ox + offx, oy + offy, offz, 0, instance->level[level]->music_artist);
	oy += 24;
}

void wl_game_music_render(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float a;

	wl_game_level_bg_render(data);
	if(instance->state == WL_STATE_MUSIC && !instance->game.music_done)
	{
		t3f_render_gui(instance->menu[WL_MENU_MUSIC_OVERLAY], T3F_GUI_USER_FLAG);
		t3f_render_gui(instance->menu[WL_MENU_MUSIC_OVERLAY], 0);
	}

	/* render song info */
	wl_game_render_music_static_info(0.0, 0.0, 0.0, al_map_rgba_f(0.0, 0.0, 0.0, 0.25), data);
	wl_game_render_music_static_info(0.0, 0.0, -8.0, WL_LIFE_FILL_COLOR, data);
	if(instance->game.tick > 0)
	{
		a = (float)instance->game.tick / WL_LEVEL_FADE_FRAMES;
		wl_game_render_music_info(1, 0.0, 0.0, 0.0, al_map_rgba_f(0.0, 0.0, 0.0, 0.25), a, data);
		wl_game_render_music_info(0, 0.0, 0.0, 0.0, al_map_rgba_f(0.0, 0.0, 0.0, 0.25), 1.0 - a, data);
		wl_game_render_music_info(1, 0.0, 0.0, -8.0, t3f_color_white, a, data);
		wl_game_render_music_info(0, 0.0, 0.0, -8.0, t3f_color_white, 1.0 - a, data);
	}
	else
	{
		wl_game_render_music_info(0, 0.0, 0.0, 0.0, al_map_rgba_f(0.0, 0.0, 0.0, 0.25), 1.0, data);
		wl_game_render_music_info(0, 0.0, 0.0, -8.0, t3f_color_white, 1.0, data);
	}
	if(instance->game.music_done)
	{
		al_draw_filled_rectangle(0.0, 0.0, WL_VIRTUAL_DISPLAY_WIDTH, WL_VIRTUAL_DISPLAY_HEIGHT, al_map_rgba_f(0.0, 0.0, 0.0, instance->game.music_fade_level));
	}
}

void wl_game_music_quit_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	wl_game_level_bg_logic(data);

	/* add time to profile */
	instance->profile->time++;

	if(instance->game.tick > 0)
	{
		instance->game.tick--;
	}

	/* see if fade is done and exit game if so */
	if(!instance->effects_data->fading)
	{
		wl_game_exit(data);
		wl_play_music(instance->music_queue, "data/music/theme.ogg");
		wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), al_map_rgba_f(0.0, 0.0, 0.0, 0.0), 60);
	}
}
