#include "t3f/sound.h"
#include "t3f/rng.h"
#include "modules/data.h"
#include "modules/achievements.h"
#include "modules/input.h"
#include "instance.h"
#include "game.h"
#include "board.h"
#include "game_intro.h"
#include "game_hazard.h"
#include "game_level.h"

bool wl_player_needs_tutorial(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	/* bring up the relevant tutorial */
	if(instance->want_tutorial)
	{
		return true;
	}
	else if(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_SKIP)
	{
		return false;
	}
	if(!instance->game.player[0].watched_tutorial)
	{
		if(!(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_BASIC))
		{
			return true;
		}
		else
		{
			if(instance->game.mode >= WL_GAME_MODE_STORY && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_STORY))
			{
				return true;
			}
			else if(instance->game.mode == WL_GAME_MODE_CASUAL && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_CASUAL))
			{
				return true;
			}
			else if(instance->game.mode == WL_GAME_MODE_PUZZLE && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_PUZZLE))
			{
				return true;
			}
			else if(instance->game.mode == WL_GAME_MODE_GURU && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_GURU))
			{
				return true;
			}
		}
	}
	return false;
}

static void wl_player_use_click(WL_PLAYER_CLICK_QUEUE * qp)
{
	int i;

	if(qp->clicks > 0)
	{
		for(i = 0; i < qp->clicks - 1; i++)
		{
			memcpy(&qp->click[i], &qp->click[i + 1], sizeof(WL_PLAYER_CLICK));
		}
		qp->clicks--;
	}
}

void wl_game_player_add_word_to_list(int word, int points, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;
	bool top = false;

	/* update main word list */
	instance->game.player[instance->game.current_player].word_list[instance->game.player[0].word_list_count] = word;
	instance->game.player[instance->game.current_player].word_list_count++;
	instance->game.player[instance->game.current_player].word_list_counter = 24;

	/* update top word list */
	for(i = 0; i < instance->game.player[instance->game.current_player].top_word_count; i++)
	{
		if(wl_game_figure_total_points(instance->dictionary, instance->game.player[instance->game.current_player].top_word[i]) < points)
		{
			instance->game.player[instance->game.current_player].top_word_spot = i;
			instance->game.player[instance->game.current_player].top_word_index = word;
			top = true;
			if(i == 0)
			{
				t3f_play_sample(instance->sample[WL_SAMPLE_BEST_WORD], 1.0, 0.0, 1.0);
			}
			break;
		}
	}

	if(instance->game.player[instance->game.current_player].top_word_count < WL_MAX_TOP_WORDS)
	{
		/* add word to end of list if it hasn't already been inserted */
		if(!top)
		{
			instance->game.player[instance->game.current_player].top_word_spot = instance->game.player[0].top_word_count;
			instance->game.player[instance->game.current_player].top_word_index = word;
			instance->game.player[instance->game.current_player].top_word[instance->game.player[instance->game.current_player].top_word_count] = word;
		}
		instance->game.player[instance->game.current_player].top_word_count++;
		top = true;
	}

	/* actually insert word */
	if(top)
	{
		for(i = WL_MAX_TOP_WORDS - 1; i > instance->game.player[instance->game.current_player].top_word_spot; i--)
		{
			instance->game.player[instance->game.current_player].top_word[i] = instance->game.player[instance->game.current_player].top_word[i - 1];
		}
		instance->game.player[instance->game.current_player].top_word[instance->game.player[instance->game.current_player].top_word_spot] = instance->game.player[instance->game.current_player].top_word_index;
	}

	/* if we got a top word, start the counter */
	if(top)
	{
		/* don't start counter if we are using the last spot since that is just
		 * a buffer for words falling off the list */
		if(instance->game.player[instance->game.current_player].top_word_spot < WL_MAX_TOP_WORDS - 1)
		{
			instance->game.player[instance->game.current_player].top_word_counter = 24;
		}
	}
}

static void wl_game_player_activate_guru(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.guru_state = 2;
	instance->game.guru_timer = 0;
	instance->game.guru_x = instance->game.player[0].sprite.x;
	instance->game.guru_y = instance->game.player[0].sprite.y;
	instance->game.guru_z = 64.0;
	instance->game.guru_vz = 0.0;
	instance->game.guru_color = al_map_rgba_f(0.0, 0.0, 0.0, 0.0);
	instance->game.guru_active = true;
	instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = '0';
}

void wl_game_player_add_life(float amount, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.player[0].life += amount;
	if(instance->game.player[0].life > WL_MAX_LIFE)
	{
		instance->game.player[0].life = WL_MAX_LIFE;
	}
	instance->game.player[0].lost = false;
}

/* subtract life, only make player lose if life is already 0.0 */
void wl_game_player_subtract_life(float amount, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	/* subtract life until it reaches 0.0 */
	if(instance->game.player[0].life > 0.0)
	{
		instance->game.player[0].life -= amount;
		if(instance->game.player[0].life < 0.0)
		{
			instance->game.player[0].life = 0.0;
		}
		else if(instance->game.player[0].life < 1.0)
		{
			instance->game.player[0].life = 1.0;
		}
	}

	/* if we are already at 0.0, lose */
	else
	{
		instance->game.player[0].lost = true;
	}
}

int wl_game_player_pos_to_move(int x, int y)
{
	return y * WL_MAX_BOARD_WIDTH + x;
}

int wl_game_player_move_to_x(int pos)
{
	return pos % WL_MAX_BOARD_WIDTH;
}

int wl_game_player_move_to_y(int pos)
{
	return pos / WL_MAX_BOARD_WIDTH;
}

static void wl_game_player_jump(int x, int y, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.player[0].sprite.vx = ((float)WL_BOARD_TILE_SIZE * (float)(x - instance->game.player[0].x)) / 21.0;
	instance->game.player[0].sprite.vy = ((float)WL_BOARD_TILE_SIZE * (float)(y - instance->game.player[0].y)) / 21.0;
	instance->game.player[0].sprite.vz = -5.0;
	instance->game.player[0].x = x;
	instance->game.player[0].y = y;
	instance->game.player[0].state = WL_PLAYER_STATE_JUMP;
	instance->game.player[0].sprite.state = WL_PLAYER_SPRITE_STATE_NORMAL;
	instance->game.player[0].sprite.tick = 0;
	t3f_play_sample(instance->sample[WL_SAMPLE_JUMP], 1.0, 0.0, 1.0);
}

static void wl_game_player_replace_letter(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->game.player[0].word_pos > 0)
	{
		instance->game.board[instance->game.player[0].y][instance->game.player[0].x].type = 2;
		instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = instance->game.player[0].word[instance->game.player[0].word_pos - 1];
		instance->game.player[0].word_pos--;
		instance->game.player[0].word[instance->game.player[0].word_pos] = '\0';
		if(instance->game.player[0].word_pos > 0 && instance->game.player[0].word[instance->game.player[0].word_pos - 1] == 'Q')
		{
			instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = 'Q';
			instance->game.player[0].word_pos--;
			instance->game.player[0].word[instance->game.player[0].word_pos] = '\0';
		}
	}
	else
	{
		instance->game.board[instance->game.player[0].y][instance->game.player[0].x].type = 3;
		instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = ' ';
	}
}

static void wl_game_player_use_valid_word(int w, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int s, m;
	char buf[16];
	int old_progress;

	instance->game.player[instance->game.current_player].words_this_level++;
	wl_generate_board_particles(al_map_rgb_f(1.0, 1.0, 1.0), instance->game.player[0].x, instance->game.player[0].y, 10, data);
	if(instance->game.player[0].powerups & WL_POWERUP_HOURGLASS)
	{
		instance->game.player[0].powerups &= ~WL_POWERUP_HOURGLASS;
	}
	s = wl_game_figure_total_points(instance->dictionary, w);
	if(s >= 3000)
	{
		m = 5;
		t3f_play_sample(instance->sample[WL_SAMPLE_AMAZING_WORD], 1.0, 0.0, 1.0);
		instance->game.player[0].sprite.state = WL_PLAYER_SPRITE_STATE_HAPPY;
		instance->game.player[0].sprite.tick = 0;
		wl_add_message(&instance->word_messages, WL_MESSAGE_TYPE_FLOAT_UP, "Amazing Word!", instance->font[WL_FONT_ASCII], 0, data);
	}
	else if(s >= 2000)
	{
		m = 3;
		t3f_play_sample(instance->sample[WL_SAMPLE_GOOD_WORD], 1.0, 0.0, 1.0);
		wl_add_message(&instance->word_messages, WL_MESSAGE_TYPE_FLOAT_UP, "Great Word!", instance->font[WL_FONT_ASCII], 0, data);
	}
	else if(s >= 1000)
	{
		m = 2;
		t3f_play_sample(instance->sample[WL_SAMPLE_GOOD_WORD], 1.0, 0.0, 1.0);
	}
	else
	{
		m = 1;
		t3f_play_sample(instance->sample[WL_SAMPLE_WORD], 1.0, 0.0, 1.0);
	}
	sprintf(buf, "%d", s);
	wl_add_message(&instance->word_messages, WL_MESSAGE_TYPE_FLOAT_UP_2, buf, instance->font[WL_FONT_ASCII], 0, data);
	instance->game.player[instance->game.current_player].score += s;
	instance->game.player[instance->game.current_player].word_score = instance->game.player[instance->game.current_player].score;
	if(instance->game.uses_life)
	{
		wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_NEAR_DEATH_VERBOSITY, data);
		wl_game_player_add_life(strlen(instance->game.player[0].word) * m, data);
	}
	if(instance->game.uses_level_progress)
	{
		old_progress = instance->game.player[0].progress;
		instance->game.player[0].progress += strlen(instance->game.player[0].word) * m;
		if(old_progress % WL_LEVEL_UP > instance->game.player[0].progress % WL_LEVEL_UP)
		{
			t3f_play_sample(instance->sample[WL_SAMPLE_LIFE_UP], 1.0, 0.0, 1.0);
			wl_game_player_add_life(WL_MAX_LIFE / 4.0, data);
			wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_TO_HEALTH, data);
		}
		if(instance->game.player[0].progress > instance->profile->progress[instance->game.player[0].story])
		{
			instance->profile->progress[instance->game.player[0].story] = instance->game.player[0].progress;
		}
	}
	else if(instance->game.mode == WL_GAME_MODE_CASUAL)
	{
		instance->game.player[0].progress += WL_LEVEL_UP / 10;
	}
	if(instance->game.mode != WL_GAME_MODE_PUZZLE)
	{
		wl_board_generate_new_letters(data);
		wl_board_logic(data);
	}
	else
	{
		if(!(instance->game.player[0].x == WL_BOARD_CENTER_X && instance->game.player[0].y == WL_BOARD_CENTER_Y) && (instance->state != WL_STATE_GAME_TUTORIAL || instance->tutorial_step >= 18))
		{
			instance->game.board[WL_BOARD_CENTER_Y][WL_BOARD_CENTER_X].letter = '@';
		}
	}
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_RARE_WORD, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_WORDSMITH, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_SUPERCALIFRAGILISTIC, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_EPONYMOUS, data);
	if(instance->game.mode == WL_GAME_MODE_GURU)
	{
		wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_A_GURU_OF_GURUS, data);
	}
	instance->game.player[0].word[0] = '\0';
	instance->game.player[0].word_pos = 0;
	instance->game.player[0].words++;
	instance->game.player[0].move[0] = wl_game_player_pos_to_move(instance->game.player[0].x, instance->game.player[0].y);
	instance->game.player[0].moves = 1;
	wl_game_player_add_word_to_list(w, s, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_HUNDREDFOLD, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_FOUL_MOUTH, data);
	wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_TAKING_IT_IN_STRIDE, data);
	if(instance->game.mode == WL_GAME_MODE_CASUAL)
	{
		wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_CONSISTENTLY_WORDY, data);
	}

	/* add word to profile */
	instance->profile->word[instance->profile->words] = w;
	instance->profile->words++;
}

static void wl_game_player_use_invalid_word(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	wl_generate_board_particles(al_map_rgb_f(1.0, 0.0, 0.0), instance->game.player[0].x, instance->game.player[0].y, 10, data);
	if(instance->game.uses_life)
	{
		wl_game_player_subtract_life(strlen(instance->game.player[0].word), data);
	}
	instance->game.player[0].sprite.state = WL_PLAYER_SPRITE_STATE_SAD;
	instance->game.player[0].sprite.tick = 0;
	instance->game.player[0].sprite.timer = 30;
	wl_board_generate_new_letters(data);
	wl_board_logic(data);
	instance->game.player[0].word[0] = '\0';
	instance->game.player[0].word_pos = 0;
	instance->game.player[0].powerups = 0;
	instance->game.player[0].move[0] = wl_game_player_pos_to_move(instance->game.player[0].x, instance->game.player[0].y);
	instance->game.player[0].moves = 1;

	t3f_play_sample(instance->sample[WL_SAMPLE_BAD_WORD], 1.0, 0.0, 1.0);

	/* add wrong guess to profile */
	instance->profile->wrongs++;
}

static void wl_game_player_use_word(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(strlen(instance->game.player[0].word) > 2)
	{
		instance->game.player[0].word_index = wl_word_in_dictionary(instance->dictionary, instance->game.player[0].word);
		if(instance->game.player[0].word_index >= 0)
		{
			wl_game_player_use_valid_word(instance->game.player[0].word_index, data);
		}
		else
		{
			if(instance->game.uses_invalid)
			{
				wl_game_player_use_invalid_word(data);
			}
		}
	}
	else if(strlen(instance->game.player[instance->game.current_player].word) > 0)
	{
		if(instance->game.uses_invalid)
		{
			wl_game_player_use_invalid_word(data);
		}
	}
	if(instance->game.guru_active)
	{
		instance->game.guru_state = 3;
		instance->game.player[0].powerups &= ~WL_POWERUP_GURU;
		instance->game.player[0].counter = 0;
	}
	instance->game.player[0].word_exists = false;
	if(instance->game.mode == WL_GAME_MODE_PARTY)
	{
		instance->game.current_player++;
		if(instance->game.current_player >= instance->game.players)
		{
			instance->game.current_player = 0;
			instance->game.player[0].progress += WL_LEVEL_UP / 10;
		}
		if(instance->game.player[0].progress < WL_LEVEL_UP)
		{
			wl_add_message(&instance->game_messages, WL_MESSAGE_TYPE_SCROLLER, instance->profile->name, instance->font[WL_FONT_32_OUTLINE], 0, data);
		}
		else
		{
			instance->game.state = WL_GAME_STATE_DONE;
		}
	}
}

static void wl_game_player_move(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float life_rate;

	/* put home space down in puzzle mode when moving away from center */
	if(instance->game.mode == WL_GAME_MODE_PUZZLE && instance->game.player[0].x == WL_BOARD_CENTER_X && instance->game.player[0].y == WL_BOARD_CENTER_Y && instance->game.player[0].words > 0 && (instance->state != WL_STATE_GAME_TUTORIAL || instance->tutorial_step >= 18))
	{
		instance->game.board[WL_BOARD_CENTER_Y][WL_BOARD_CENTER_X].letter = '@';
	}

	/* put back letter */
	else if(instance->game.board[instance->game.player[0].click_queue.click[0].by][instance->game.player[0].click_queue.click[0].bx].letter == ' ')
	{
		wl_game_player_replace_letter(data);

		/* add undo to profile */
		instance->profile->undos++;
	}

	/* remove letter */
	else
	{
		if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter != '@')
		{
			instance->game.board[instance->game.player[0].y][instance->game.player[0].x].type = 3;
			instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = ' ';
		}
	}

	/* make player jump */
	wl_game_player_jump(instance->game.player[0].click_queue.click[0].bx, instance->game.player[0].click_queue.click[0].by, data);
	if(instance->game.uses_life && !(instance->game.player[0].powerups & WL_POWERUP_HOURGLASS))
	{
		if(instance->game.mode == WL_GAME_MODE_GURU)
		{
			life_rate = WL_LIFE_DECREASE_GURU_MODE;
		}
		else
		{
			life_rate = instance->level_list->entry[instance->game.player[0].level].life_decrease;
		}
		wl_game_player_subtract_life(life_rate, data);
	}
}

static bool wl_mouse_in_board(float mouse_x, float mouse_y)
{
	if(mouse_x - WL_BOARD_X < 0)
	{
		return false;
	}
	else if(mouse_x - WL_BOARD_X >= WL_MAX_BOARD_WIDTH * WL_BOARD_TILE_SIZE)
	{
		return false;
	}
	else if(mouse_y - WL_BOARD_Y < 0)
	{
		return false;
	}
	else if(mouse_y - WL_BOARD_Y >= WL_MAX_BOARD_HEIGHT * WL_BOARD_TILE_SIZE)
	{
		return false;
	}
	return true;
}

static void _update_hover_position(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->game.player[0].hover_x < 0 || instance->game.player[0].hover_y < 0)
	{
		instance->game.player[0].hover_x = instance->game.player[0].x;
		instance->game.player[0].hover_y = instance->game.player[0].y;
	}
}

static void _handle_manual_cursor_movement(bool up, bool down, bool left, bool right, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(up)
	{
		if(instance->game.player[0].hover_y > 0)
		{
			instance->game.player[0].hover_y--;
		}
	}
	if(down)
	{
		if(instance->game.player[0].hover_y < WL_MAX_BOARD_HEIGHT - 1)
		{
			instance->game.player[0].hover_y++;
		}
	}
	if(left)
	{
		if(instance->game.player[0].hover_x > 0)
		{
			instance->game.player[0].hover_x--;
		}
	}
	if(right)
	{
		if(instance->game.player[0].hover_x < WL_MAX_BOARD_WIDTH - 1)
		{
			instance->game.player[0].hover_x++;
		}
	}
}

static void _handle_cursor_movement(int mouse_x, int mouse_y, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(wl_mouse_in_board(mouse_x, mouse_y))
	{
		instance->game.player[0].hover_x = (mouse_x - WL_BOARD_X) / WL_BOARD_TILE_SIZE;
		instance->game.player[0].hover_y = (mouse_y - WL_BOARD_Y) / WL_BOARD_TILE_SIZE;
		instance->game.player[0].board_x = (mouse_x - WL_BOARD_X) / WL_BOARD_TILE_SIZE;
		instance->game.player[0].board_y = (mouse_y - WL_BOARD_Y) / WL_BOARD_TILE_SIZE;
	}
	else
	{
		instance->game.player[0].hover_x = -1;
		instance->game.player[0].hover_y = -1;
		instance->game.player[0].board_x = -1;
		instance->game.player[0].board_y = -1;
	}
}

static void _handle_click(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->game.player[0].click_queue.clicks < WL_PLAYER_MAX_CLICKS)
	{
		if(instance->game.player[0].hover_x >= 0 && instance->game.player[0].hover_y >= 0)
		{
			instance->game.player[0].click_queue.click[instance->game.player[0].click_queue.clicks].bx = instance->game.player[0].hover_x;
			instance->game.player[0].click_queue.click[instance->game.player[0].click_queue.clicks].by = instance->game.player[0].hover_y;
			instance->game.player[0].click_queue.clicks++;
		}
	}
}

static void _handle_keyboard_controls(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_update_hover_position(data);
	_handle_manual_cursor_movement(instance->input_state.up.pressed, instance->input_state.down.pressed, instance->input_state.left.pressed, instance->input_state.right.pressed, data);
	if(instance->input_state.action.pressed)
	{
		_handle_click(data);
	}
}

static void _handle_gamepad_controls(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_update_hover_position(data);
	_handle_manual_cursor_movement(instance->input_state.up.pressed, instance->input_state.down.pressed, instance->input_state.left.pressed, instance->input_state.right.pressed, data);
	if(instance->input_state.action.pressed)
	{
		_handle_click(data);
	}
}

static void _handle_touch_controls(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int mouse_x = -10000;
	int mouse_y = -10000;
	bool button = false;
	bool update_cursor = false;

	/* store touch input in local variables */
	if(instance->input_state.action.held)
	{
		mouse_x = t3f_get_touch_x(1);
		mouse_y = t3f_get_touch_y(1);
		update_cursor = true;
	}
	else if(instance->input_state.action.released)
	{
		mouse_x = t3f_get_touch_x(1);
		mouse_y = t3f_get_touch_y(1);
		update_cursor = true;

		/* only alter touch state if we are clicking inside the board area */
		if(wl_mouse_in_board(mouse_x, mouse_y))
		{
			button = true;
		}
	}
	if(update_cursor)
	{
		_handle_cursor_movement(mouse_x, mouse_y, data);
	}
	if(button)
	{
		_handle_click(data);
	}
}

static void _handle_mouse_controls(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_handle_cursor_movement(instance->input_state.pointer.x, instance->input_state.pointer.y, data);
	if(instance->input_state.action.pressed)
	{
		_handle_click(data);
	}
}

static void _verify_range(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	/* see if we are within range for the current position */
	if(instance->game.player[0].hover_x >= 0 && instance->game.player[0].hover_y >= 0 && abs(instance->game.player[0].x - instance->game.player[0].hover_x) < instance->game.player[0].range && abs(instance->game.player[0].y - instance->game.player[0].hover_y) < instance->game.player[0].range)
	{
		instance->game.player[0].in_range = true;
	}
	else
	{
		instance->game.player[0].in_range = false;
	}
}

static void _process_click_queue(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	switch(instance->game.player[0].state)
	{
		case WL_PLAYER_STATE_NORMAL:
		{
			/* process next click */
			if(instance->game.player[0].click_queue.clicks > 0)
			{
				/* see if the click is in range */
				if((abs(instance->game.player[0].x - instance->game.player[0].click_queue.click[0].bx) < instance->game.player[0].range && abs(instance->game.player[0].y - instance->game.player[0].click_queue.click[0].by) < instance->game.player[0].range) && !(instance->game.board[instance->game.player[0].click_queue.click[0].by][instance->game.player[0].click_queue.click[0].bx].flags & WL_BOARD_FLAG_BLOCKED))
				{

					/* clicked on self to use current word */
					if(instance->game.player[0].click_queue.click[0].bx == instance->game.player[0].x && instance->game.player[0].click_queue.click[0].by == instance->game.player[0].y)
					{
						wl_game_player_use_word(data);
					}
					/* clicked space other than player space */
					else
					{
						wl_game_player_move(data);
					}
				}

				/* if click is outside of range, remove all clicks from the queue,
				 * player probably wanted a letter that was out of range */
				else
				{
					instance->game.player[0].click_queue.clicks = 0;
				}
			}
			wl_player_use_click(&instance->game.player[0].click_queue);
			break;
		}
	}
}

void wl_game_player_control(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	switch(instance->input_type)
	{
		case WL_INPUT_TYPE_MOUSE:
		{
			_handle_mouse_controls(data);
			break;
		}
		case WL_INPUT_TYPE_TOUCH:
		{
			_handle_touch_controls(data);
			break;
		}
		case WL_INPUT_TYPE_KEYBOARD:
		{
			_handle_keyboard_controls(data);
			break;
		}
		case WL_INPUT_TYPE_GAMEPAD:
		{
			_handle_gamepad_controls(data);
			break;
		}
	}

	_verify_range(data);
	_process_click_queue(data);
}

void wl_game_player_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int base_range = instance->game.base_range;
	int i, p;
	bool might_swoosh = false;

	if(instance->state != WL_STATE_GAME && instance->state != WL_STATE_GAME_TUTORIAL)
	{
		base_range = 0;
	}
	if(instance->game.mode == WL_GAME_MODE_PUZZLE)
	{
		base_range = WL_MAX_BOARD_WIDTH;
	}

	/* adjust range if needed */
	if(instance->game.player[0].powerups & WL_POWERUP_GURU)
	{
		if(instance->game.player[0].range < WL_MAX_BOARD_WIDTH)
		{
			instance->game.player[0].counter++;
			if(instance->game.player[0].counter % 4 == 0)
			{
				instance->game.player[0].range++;
			}
		}
	}
	else
	{
		if(instance->game.player[0].range > base_range)
		{
			instance->game.player[0].counter++;
			if(instance->game.player[0].counter % 4 == 0)
			{
				instance->game.player[0].range--;
			}
		}
		else if(instance->game.player[0].range < base_range)
		{
			instance->game.player[0].counter++;
			if(instance->game.player[0].counter % 4 == 0)
			{
				instance->game.player[0].range++;
			}
		}
	}

	switch(instance->game.player[0].state)
	{
		case WL_PLAYER_STATE_ENTER:
		{
			if(instance->game.player[0].sprite.z < WL_SPRITE_SWOOSH_HEIGHT)
			{
				might_swoosh = true;
			}
			instance->game.player[0].sprite.z += instance->game.player[0].sprite.vz;
			if(might_swoosh && instance->game.player[0].sprite.z >= WL_SPRITE_SWOOSH_HEIGHT)
			{
				t3f_play_sample(instance->sample[WL_SAMPLE_SWOOSH], 1.0, 0.0, 1.0);
			}
			if(instance->game.player[0].sprite.z >= 0)
			{
				t3f_play_sample(instance->sample[WL_SAMPLE_LAND], 1.0, 0.0, 1.0);
				instance->game.player[0].sprite.z = 0;
				instance->game.player[0].state = WL_PLAYER_STATE_NORMAL;

				if(wl_player_needs_tutorial(data))
				{
					/* bring up the relevant tutorial */
					if(instance->want_tutorial || !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_BASIC))
					{
						instance->tutorial_step = 0;
						wl_game_setup_type(data);
						instance->game.uses_life = false;
						instance->game.uses_level_progress = false;
						instance->state = WL_STATE_GAME_TUTORIAL;
					}
					else
					{
						if(instance->game.mode >= WL_GAME_MODE_STORY && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_STORY))
						{
							instance->tutorial_step = 14;
							instance->state = WL_STATE_GAME_TUTORIAL;
						}
						else if(instance->game.mode == WL_GAME_MODE_CASUAL && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_CASUAL))
						{
							instance->tutorial_step = 14;
							instance->state = WL_STATE_GAME_TUTORIAL;
						}
						else if(instance->game.mode == WL_GAME_MODE_PUZZLE && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_PUZZLE))
						{
							instance->tutorial_step = 14;
							instance->state = WL_STATE_GAME_TUTORIAL;
						}
						else if(instance->game.mode == WL_GAME_MODE_GURU && !(instance->profile->flags & WL_PROFILE_FLAG_TUTORIAL_GURU))
						{
							instance->tutorial_step = 14;
							instance->state = WL_STATE_GAME_TUTORIAL;
						}
					}
				}
			}
			break;
		}
		case WL_PLAYER_STATE_EXIT:
		{
			if(instance->game.player[0].sprite.z >= WL_SPRITE_SWOOSH_HEIGHT)
			{
				might_swoosh = true;
			}
			instance->game.player[0].sprite.z -= WL_TERMINAL_VELOCITY;
			if(might_swoosh && instance->game.player[0].sprite.z < WL_SPRITE_SWOOSH_HEIGHT)
			{
				t3f_play_sample(instance->sample[WL_SAMPLE_SWOOSH], 1.0, 0.0, 1.0);
			}
			break;
		}
		case WL_PLAYER_STATE_NORMAL:
		{
			instance->game.player[0].sprite.x = instance->game.player[0].x * WL_BOARD_TILE_SIZE;
			instance->game.player[0].sprite.y = instance->game.player[0].y * WL_BOARD_TILE_SIZE;
			instance->game.player[0].sprite.z = 0;
			break;
		}
		case WL_PLAYER_STATE_JUMP:
		{
			instance->game.player[0].sprite.x += instance->game.player[0].sprite.vx;
			instance->game.player[0].sprite.y += instance->game.player[0].sprite.vy;
			instance->game.player[0].sprite.z += instance->game.player[0].sprite.vz;
			if(instance->game.player[0].sprite.z >= 0.0)
			{
				t3f_play_sample(instance->sample[WL_SAMPLE_LAND], 1.0, 0.0, 1.0);
				if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter == '?')
				{
					t3f_play_sample(instance->sample[WL_SAMPLE_GURU], 1.0, 0.0, 1.0);
					instance->game.player[0].powerups |= WL_POWERUP_GURU;

					/* activate guru */
					wl_game_player_activate_guru(data);
				}
				else if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter == '!')
				{
					t3f_play_sample(instance->sample[WL_SAMPLE_LIFE_UP], 1.0, 0.0, 1.0);
					instance->game.player[0].life = WL_MAX_LIFE;
					instance->game.player[0].lost = false;
					instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = '0';
				}
				else if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter == '_')
				{
					t3f_play_sample(instance->sample[WL_SAMPLE_HOURGLASS], 1.0, 0.0, 1.0);
					instance->game.player[0].powerups |= WL_POWERUP_HOURGLASS;
					instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = '0';
				}
				else if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter == '@' && instance->game.mode == WL_GAME_MODE_PUZZLE)
				{
					if(instance->state != WL_STATE_GAME_TUTORIAL)
					{
						wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_PUZZLE_SOLVER, data);
						wl_game_end(data);
					}
				}
				else if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter == 'Q')
				{
					instance->game.player[0].word[instance->game.player[0].word_pos] = 'Q';
					instance->game.player[0].word_pos++;
					instance->game.player[0].word[instance->game.player[0].word_pos] = '\0';
					instance->game.player[0].word[instance->game.player[0].word_pos] = 'U';
					instance->game.player[0].word_pos++;
					instance->game.player[0].word[instance->game.player[0].word_pos] = '\0';
					instance->game.player[0].move[instance->game.player[0].moves] = wl_game_player_pos_to_move(instance->game.player[0].x, instance->game.player[0].y);
					instance->game.player[0].moves++;
				}
				else if(instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter != ' ' && instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter != '0')
				{
					instance->game.player[0].word[instance->game.player[0].word_pos] = instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter;
					instance->game.player[0].word_pos++;
					instance->game.player[0].word[instance->game.player[0].word_pos] = '\0';
					instance->game.player[0].move[instance->game.player[0].moves] = wl_game_player_pos_to_move(instance->game.player[0].x, instance->game.player[0].y);
					instance->game.player[0].moves++;
				}
				instance->game.board[instance->game.player[0].y][instance->game.player[0].x].letter = ' ';
				/* add move to profile */
				instance->profile->moves++;

				/* we set state to EXIT in some cases so only revert to NORMAL if we are still JUMPing */
				if(instance->game.player[0].state == WL_PLAYER_STATE_JUMP)
				{
					instance->game.player[0].state = WL_PLAYER_STATE_NORMAL;
				}
				if(strlen(instance->game.player[0].word) >= 3 && wl_word_in_dictionary(instance->dictionary, instance->game.player[0].word) >= 0)
				{
					instance->game.player[0].word_exists = true;
				}
				else
				{
					instance->game.player[0].word_exists = false;
				}
			}
			instance->game.player[0].sprite.vz += 0.5;
			break;
		}
	}

	/* animate HUD */
	if(instance->game.player[0].life_indicator < instance->game.player[0].life)
	{
		instance->game.player[0].life_indicator += 1.0;
		if(instance->game.player[0].life_indicator > instance->game.player[0].life)
		{
			instance->game.player[0].life_indicator = instance->game.player[0].life;
		}
	}
	else if(instance->game.player[0].life_indicator > instance->game.player[0].life)
	{
		instance->game.player[0].life_indicator -= 1.0;
		if(instance->game.player[0].life_indicator < instance->game.player[0].life)
		{
			instance->game.player[0].life_indicator = instance->game.player[0].life;
		}
	}
	if(instance->game.player[0].progress_indicator < instance->game.player[0].progress)
	{
		instance->game.player[0].progress_indicator++;
		instance->game.player[0].made_progress = true;
	}
	else if(instance->game.player[0].progress_indicator > instance->game.player[0].progress)
	{
		instance->game.player[0].progress_indicator--;
	}

	/* life meter graphical effect */
	if(!(instance->game.player[0].powerups & WL_POWERUP_HOURGLASS))
	{
		/* only process fading when fade level is > 0.0 to prevent triggering
		 * extra particle generation */
		if(instance->game.player[0].life_fade > 0.0)
		{
			if(instance->game.player[0].life_fade >= 1.0)
			{
				/* generate steam particles when color fade starts */
				for(i = 0; i < 32; i++)
				{
					p = wl_add_particle(instance->particle_list[1], WL_PARTICLE_TYPE_STEAM, instance->animation[WL_ANIMATION_PARTICLE_4], al_map_rgba_f(1.0, 1.0, 1.0, 0.75), WL_LIFE_BAR_X + t3f_drand(&instance->rng_state) * instance->bitmap[WL_BITMAP_LIFE_FILL]->target_width, WL_LIFE_BAR_Y + t3f_drand(&instance->rng_state) * instance->bitmap[WL_BITMAP_LIFE_FILL]->target_height, 0.0, 1.0, 60, 0);
					instance->particle_list[1]->particle[p].vz = -2.5;
					instance->particle_list[1]->particle[p].vy = -t3f_drand(&instance->rng_state);
					instance->particle_list[1]->particle[p].vx = t3f_drand(&instance->rng_state) - 0.5;
					t3f_play_sample(instance->sample[WL_SAMPLE_THAW], 1.0, 0.0, 1.0);
				}
			}
			instance->game.player[0].life_fade -= 4.0 * (1.0 / 60.0);
			if(instance->game.player[0].life_fade <= 0.0)
			{
				instance->game.player[0].life_fade = 0.0;
			}
		}
	}
	else
	{
		/* only process fading when fade level is < 1.0 to prevent triggering
		 * extra particle generation */
		if(instance->game.player[0].life_fade < 1.0)
		{
			instance->game.player[0].life_fade += 4.0 * (1.0 / 60.0);
			if(instance->game.player[0].life_fade >= 1.0)
			{
				instance->game.player[0].life_fade = 1.0;

				/* generate ice particles when color fade completes */
				for(i = 0; i < 32; i++)
				{
					p = wl_add_particle(instance->particle_list[1], WL_PARTICLE_TYPE_ICE, instance->animation[WL_ANIMATION_PARTICLE_3], al_map_rgba_f(0.9, 0.9, 1.0, 0.75), WL_LIFE_BAR_X + t3f_drand(&instance->rng_state) * instance->bitmap[WL_BITMAP_LIFE_FILL]->target_width, WL_LIFE_BAR_Y + t3f_drand(&instance->rng_state) * instance->bitmap[WL_BITMAP_LIFE_FILL]->target_height, 0.0, 1.0, 60, 0);
					instance->particle_list[1]->particle[p].vz = -0.5;
					instance->particle_list[1]->particle[p].vy = 0.0;
					instance->particle_list[1]->particle[p].vx = t3f_drand(&instance->rng_state) - 0.5;
					t3f_play_sample(instance->sample[WL_SAMPLE_FREEZE], 1.0, 0.0, 1.0);
				}
			}
		}
	}

	/* animate word lists */
	if(instance->game.player[0].word_list_counter > 0)
	{
		instance->game.player[0].word_list_counter--;
	}
	if(instance->game.player[0].top_word_counter > 0)
	{
		instance->game.player[0].top_word_counter--;
	}

	/* play low life warning sound */
	if(instance->game.uses_life && instance->state == WL_STATE_GAME)
	{
		if(instance->game.player[0].life <= 10.0 && !(instance->game.player[0].powerups & WL_POWERUP_HOURGLASS) && instance->tick % 60 == 0)
		{
			t3f_play_sample(instance->sample[WL_SAMPLE_LIFE_LOW], 1.0, 0.0, 1.0);
		}
	}
}

void wl_player_rules_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j, p;
	WL_LEVEL * lev;

	/* when the progress bar is full, generate "glass" particles and load the next level */
	if(instance->game.player[0].made_progress)
	{
		if(instance->game.player[0].progress_indicator % WL_LEVEL_UP == 0)
		{
			t3f_play_sample(instance->sample[WL_SAMPLE_SHATTER], 1.0, 0.0, 1.0);
			for(i = 0; i < instance->bitmap[WL_BITMAP_LEVEL_FILL]->target_width / instance->animation[WL_ANIMATION_PARTICLE_2]->data->frame[0]->width; i++)
			{
				for(j = 0; j < instance->bitmap[WL_BITMAP_LEVEL_FILL]->target_height / instance->animation[WL_ANIMATION_PARTICLE_2]->data->frame[0]->height; j++)
				{
					p = wl_add_particle(instance->particle_list[1], WL_PARTICLE_TYPE_GLASS, instance->animation[WL_ANIMATION_PARTICLE_2], al_map_rgba(172, 212, 255, 141), WL_LEVEL_METER_X + i * instance->animation[WL_ANIMATION_PARTICLE_2]->data->frame[0]->width, WL_LEVEL_METER_Y + j * instance->animation[WL_ANIMATION_PARTICLE_2]->data->frame[0]->height, 0.0, (float)WL_BOARD_TILE_SIZE, 20 + rand() % 20, 0);
					instance->particle_list[1]->particle[p].vx = 0.0;
					instance->particle_list[1]->particle[p].vy = t3f_drand(&instance->rng_state) * -4.0;
					instance->particle_list[1]->particle[p].vz = -(t3f_drand(&instance->rng_state) * 10.0 + 1.0);
				}
			}
			if(instance->game.mode >= WL_GAME_MODE_STORY)
			{
				if(instance->game.player[0].level < instance->level_list->story[instance->game.player[0].story].end_level)
				{
					instance->game.want_new_level++;
					wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_QUICKIE, data);
				}
				else
				{
					instance->game.finished = true;
					wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_ENLIGHTENMENT, data);
					wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_TRUE_ENLIGHTENMENT, data);
					wl_game_end(data);
				}
			}
			else if(instance->game.mode == WL_GAME_MODE_CASUAL || instance->game.mode == WL_GAME_MODE_PARTY)
			{
				wl_game_end(data);
			}
		}
		instance->game.player[0].made_progress = false;
	}

	if(instance->game.want_new_level)
	{
		/* go to next level as soon as possible */
		if(wl_level_transition_complete(data))
		{
			/* swap preloaded level to slot 0 */
			lev = instance->level[0];
			instance->level[0] = instance->level[1];
			instance->level[1] = lev;

			wl_play_level_music(instance->level[0], data);

			instance->game.player[0].level++;
			wl_update_hazard_chart(data);
			wl_add_message(&instance->game_messages, WL_MESSAGE_TYPE_SCROLLER, instance->level[0]->name, instance->font[WL_FONT_32_OUTLINE], 0, data);
			instance->game.tick = WL_LEVEL_FADE_FRAMES;
			instance->game.player[0].words_this_level = 0;
			instance->game.want_new_level--;
			instance->game.got_new_level = true;
		}
	}

	/* see if the game is over */
	if(instance->game.uses_life && instance->game.player[0].state != WL_PLAYER_STATE_ENTER)
	{
		if(instance->game.player[0].life <= 0.0 && instance->game.player[0].lost)
		{
			instance->game.player[0].life = 0.0;
			if(instance->game.player[0].state == WL_PLAYER_STATE_NORMAL && instance->state != WL_STATE_GAME_TUTORIAL)
			{
				wl_achievements_logic(instance->profile, WL_ACHIEVEMENT_LOSER, data);
				wl_game_end(data);
			}
		}
	}
}
