#include "t3f/t3f.h"
#include "t3f/resource.h"
#include "t3f/windows.h"
#include "t3f/input.h"
#include "modules/data.h"
#include "modules/video_mode.h"
#include "modules/text_entry.h"
#include "modules/effects.h"
#include "modules/music_queue.h"
#include "modules/level_list.h"
#include "modules/leaderboards.h"
#include "modules/character_list.h"
#include "modules/input.h"
#include "modules/achievements.h"
#include "modules/nine_patch.h"
#include "ui/menu_options.h"
#include "ui/menu_profile.h"
#include "instance.h"
#include "preferences.h"
#include "event.h"

void wl_logic(void * data);
void wl_render(void * data);

static int _wl_load_color = 255;
static float _wl_load_time = -1.0;

static void _wl_show_load_icon(T3F_BITMAP * bp)
{
	float current_time = al_get_time();

	#ifdef ALLEGRO_ANDROID
		t3f_process_events(false);
	#endif
	if(current_time - _wl_load_time >= 1.0)
	{
		t3f_select_view(t3f_default_view);
		al_clear_to_color(WL_BASE_COLOR);
		t3f_draw_bitmap(bp, al_map_rgba(_wl_load_color, _wl_load_color, _wl_load_color, _wl_load_color), t3f_default_view->right - bp->target_width - 8, t3f_default_view->bottom - bp->target_height - 8, 0, 0);
		_wl_load_color -= 128;
		if(_wl_load_color <= 0)
		{
			_wl_load_color = 255;
		}
		al_flip_display();
		_wl_load_time = current_time;
	}
}

bool wl_load_bitmap(int bitmap, const char * fn, int flags, float scale, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[1024];

	_wl_show_load_icon(instance->bitmap[WL_BITMAP_LOAD_ICON]);
	sprintf(buf, "%s#Size %1.1f", fn, scale);
	instance->bitmap[bitmap] = t3f_load_bitmap(buf, flags, false);
	if(!instance->bitmap[bitmap])
	{
		printf("Failed to load bitmap %d!\n", bitmap);
		return false;
	}
	return true;
}

bool wl_load_nine_patch(NINE_PATCH_BITMAP ** npbp, const char * fn, float scale)
{
	char buf[1024];

	sprintf(buf, "%s#Size %1.1f", fn, scale);
	if(np_load_nine_patch_bitmap(npbp, buf))
	{
		return true;
	}
	return false;
}

bool wl_load_tileset(int tile, const char * fn, int flags, float scale, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[1024];

	sprintf(buf, "%s#Size %1.1f", fn, scale);
	instance->tile[tile] = wl_load_tile(buf, flags);
	if(!instance->tile[tile])
	{
		printf("Failed to load tile %d!\n", tile);
		return false;
	}
	return true;
}

bool wl_load_sample(int sample, const char * fn, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->sample[sample] = al_load_sample(fn);
	if(!instance->sample[sample])
	{
		printf("Failed to load sample %d!\n", sample);
		return false;
	}
	return true;
}

bool wl_load_animation(int animation, const char * fn, float scale, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[1024];

	_wl_show_load_icon(instance->bitmap[WL_BITMAP_LOAD_ICON]);
	sprintf(buf, "%s#Size %1.1f", fn, scale);
	instance->animation[animation] = t3f_load_animation(buf, T3F_BITMAP_FLAG_PADDED, false);
	if(!instance->animation[animation])
	{
		printf("Failed to load animation %d!\n", animation);
		return false;
	}
	return true;
}

bool wl_load_font(int font, const char * fn, int size, float scale, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[1024];

	_wl_show_load_icon(instance->bitmap[WL_BITMAP_LOAD_ICON]);
	sprintf(buf, "%s#Size %1.1f", fn, scale);
	instance->font[font] = t3f_load_font(buf, T3F_FONT_TYPE_AUTO, size, 0, false);
	if(!instance->font[font])
	{
		printf("Failed to load font %d!\n", font);
		return false;
	}
	return true;
}

bool wl_load_data(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	WL_DICTIONARY * censor_dictionary = NULL;
	char fnbuf[1024] = {0};
	int i;

	instance->bitmap[WL_BITMAP_LOAD_ICON] = t3f_load_bitmap("data/graphics/core/load_icon.png", 0, false);
	if(!instance->bitmap[WL_BITMAP_LOAD_ICON])
	{
		return false;
	}

	instance->leaderboard_loader = t3f_create_object_loader();
	if(!instance->leaderboard_loader)
	{
		return false;
	}

	/* load animations */
	if(!wl_load_animation(WL_ANIMATION_GURU, "data/graphics/core/guru.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_BOMB, "data/graphics/core/bomb.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_BOMB_SHADOW, "data/graphics/core/bomb_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_LASER, "data/graphics/core/laser.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_LASER_SHADOW, "data/graphics/core/laser_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_PARTICLE, "data/graphics/core/particle.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_PARTICLE_2, "data/graphics/core/particle2.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_PARTICLE_3, "data/graphics/core/particle3.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_PARTICLE_4, "data/graphics/core/particle4.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_HEALTH, "data/graphics/core/powerup_health.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_HEALTH_SHADOW, "data/graphics/core/powerup_health_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_GURU, "data/graphics/core/powerup_guru.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_GURU_SHADOW, "data/graphics/core/powerup_guru_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_HOURGLASS, "data/graphics/core/powerup_hourglass.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_POWER_HOURGLASS_SHADOW, "data/graphics/core/powerup_hourglass_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_SPIKE, "data/graphics/core/spike.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_SPIKE_2, "data/graphics/core/spike2.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_SPIKE_SHADOW, "data/graphics/core/spike_shadow.ini", instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_animation(WL_ANIMATION_SELECTOR, "data/graphics/core/selector.ini", instance->graphics_scale, data))
	{
		return false;
	}

	/* load bitmaps */
	if(!wl_load_bitmap(WL_BITMAP_LIFE, "data/graphics/core/life.ini", T3F_BITMAP_FLAG_PADDED, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_bitmap(WL_BITMAP_LIFE_FILL, "data/graphics/core/life_fill.ini", 0, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_bitmap(WL_BITMAP_LIFE_OVERLAY, "data/graphics/core/life_overlay.ini", 0, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_bitmap(WL_BITMAP_LEVEL, "data/graphics/core/level.ini", T3F_BITMAP_FLAG_PADDED, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_bitmap(WL_BITMAP_LEVEL_FILL, "data/graphics/core/level_fill.ini", 0, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_bitmap(WL_BITMAP_POPUP, "data/graphics/core/popup.ini", 0, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_nine_patch(&instance->box_nine_patch, "data/graphics/core/box_85.ini", instance->graphics_scale))
	{
		return false;
	}
	if(!wl_load_tileset(WL_TILE_MENU_24, "data/graphics/core/tile_24.ini", WL_TILE_CONNECT_NONE | WL_TILE_CONNECT_H, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_tileset(WL_TILE_MENU_32, "data/graphics/core/tile_32.ini", WL_TILE_CONNECT_NONE | WL_TILE_CONNECT_H, instance->graphics_scale, data))
	{
		return false;
	}
	if(!wl_load_tileset(WL_TILE_BOARD, "data/graphics/core/tile_48.ini", WL_TILE_CONNECT_VH, instance->graphics_scale, data))
	{
		return false;
	}

	if(!wl_load_characters(data))
	{
		printf("failed to load characters\n");
		return false;
	}
	if(instance->character_count < 1)
	{
		printf("no characters found!\n");
		return false;
	}

	/* create atlas */
	instance->atlas[WL_ATLAS_GAME] = t3f_create_atlas(440 * instance->graphics_scale, 440 * instance->graphics_scale);
	if(instance->atlas[WL_ATLAS_GAME])
	{
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_GAME], &instance->bitmap[WL_BITMAP_POPUP]->bitmap, T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_SELECTOR], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_BOMB], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_BOMB_SHADOW], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_LASER], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_LASER_SHADOW], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_PARTICLE], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_PARTICLE_2], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_PARTICLE_3], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_PARTICLE_4], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_HEALTH], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_HEALTH_SHADOW], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_GURU], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_GURU_SHADOW], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_HOURGLASS], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_POWER_HOURGLASS_SHADOW], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_SPIKE], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_SPIKE_2], T3F_ATLAS_SPRITE);
		t3f_add_animation_to_atlas(instance->atlas[WL_ATLAS_GAME], instance->animation[WL_ANIMATION_SPIKE_SHADOW], T3F_ATLAS_SPRITE);
		for(i = 0; i < instance->character_count; i++)
		{
			wl_add_character_to_atlas(instance->character[i], instance->atlas[WL_ATLAS_GAME]);
		}
	}

	/* create HUD atlas for slight optimization */
	instance->atlas[WL_ATLAS_HUD] = t3f_create_atlas(448 * instance->graphics_scale, 380 * instance->graphics_scale);
	if(instance->atlas[WL_ATLAS_HUD])
	{
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_HUD], &instance->bitmap[WL_BITMAP_LEVEL]->bitmap, T3F_ATLAS_SPRITE);
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_HUD], &instance->bitmap[WL_BITMAP_LEVEL_FILL]->bitmap, T3F_ATLAS_SPRITE);
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_HUD], &instance->bitmap[WL_BITMAP_LIFE]->bitmap, T3F_ATLAS_SPRITE);
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_HUD], &instance->bitmap[WL_BITMAP_LIFE_FILL]->bitmap, T3F_ATLAS_SPRITE);
		t3f_add_bitmap_to_atlas(instance->atlas[WL_ATLAS_HUD], &instance->bitmap[WL_BITMAP_LIFE_OVERLAY]->bitmap, T3F_ATLAS_SPRITE);
	}

	/* load fonts */
	if(!wl_load_font(WL_FONT_24, "data/fonts/teen.ini", 24, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_24_OUTLINE, "data/fonts/teen_outline.ini", 24, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_16, "data/fonts/teen.ini", 14, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_16_OUTLINE, "data/fonts/teen_outline.ini", 14, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_32, "data/fonts/teen.ini", 40, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_32_OUTLINE, "data/fonts/teen_outline.ini", 40, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_12, "data/fonts/teen.ini", 12, instance->graphics_scale, instance))
	{
		return false;
	}
	if(!wl_load_font(WL_FONT_ASCII, "data/fonts/teen_outline.ini", 24, instance->graphics_scale, instance))
	{
		return false;
	}

	/* load sounds */
	if(!wl_load_sample(WL_SAMPLE_JUMP, "data/sounds/jump.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_LAND, "data/sounds/land.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_BAD_WORD, "data/sounds/bad_word.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_WORD, "data/sounds/word.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_GOOD_WORD, "data/sounds/good_word.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_AMAZING_WORD, "data/sounds/amazing_word.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_SHATTER, "data/sounds/shatter.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_BOMB, "data/sounds/bomb.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_SPIKE, "data/sounds/spike.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_LASER, "data/sounds/laser.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_LIFE_UP, "data/sounds/life_up.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_LIFE_LOW, "data/sounds/life_low.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_GURU, "data/sounds/guru.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_HOURGLASS, "data/sounds/hourglass.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_SWOOSH, "data/sounds/swoosh.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_HURT, "data/sounds/hurt.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_DESTROY, "data/sounds/destroy.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_BEST_WORD, "data/sounds/best_word.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_FREEZE, "data/sounds/freeze.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_THAW, "data/sounds/thaw.wav", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_MENU_TAP, "data/sounds/menu_tap.ogg", data))
	{
		return false;
	}
	if(!wl_load_sample(WL_SAMPLE_MENU_MOVE, "data/sounds/menu_move.ogg", data))
	{
		return false;
	}

	instance->dictionary = wl_load_dictionary("data/misc/dictionary.txt");
	if(!instance->dictionary)
	{
		printf("Unable to load dictionary!\n");
		return false;
	}
	instance->dictionary_checksum = t3f_checksum_file("data/misc/dictionary.txt");
	if(!instance->option[WL_OPTION_DISABLE_CENSORSHIP])
	{
		censor_dictionary = wl_load_dictionary("data/misc/censored.txt");
		if(!censor_dictionary)
		{
			printf("Unable to load censored word list, required because option 'Disable Censorship' is not set!\n");
			return false;
		}
		wl_apply_censorship_to_dictionary(instance->dictionary, censor_dictionary);
		wl_destroy_dictionary(censor_dictionary);
	}

	instance->profile = wl_load_profile(t3f_get_filename(t3f_data_path, "profile.dat", fnbuf, 1024));
	if(!instance->profile)
	{
		instance->profile = wl_create_profile();
		if(!instance->profile)
		{
			return false;
		}
	}

	instance->effects_data = wl_create_effects_data();
	if(!instance->effects_data)
	{
		return false;
	}

	instance->music_queue = wl_create_music_queue();
	if(!instance->music_queue)
	{
		return false;
	}
	wl_set_music_base_volume(instance->music_queue, (float)instance->option[WL_OPTION_MUSIC_VOLUME] / 100.0);

	instance->particle_list[0] = wl_create_particle_list(WL_MAX_PARTICLES);
	if(!instance->particle_list[0])
	{
		return false;
	}
	instance->particle_list[1] = wl_create_particle_list(WL_MAX_PARTICLES);
	if(!instance->particle_list[1])
	{
		return false;
	}

	instance->shape_lists = wl_create_shape_lists(WL_MAX_SHAPES);
	if(!instance->shape_lists)
	{
		return false;
	}

	instance->text_entry = wl_create_text_entry_data();
	if(!instance->text_entry)
	{
		return false;
	}

	i = wl_count_installed_levels();
	if(i <= 0)
	{
		return false;
	}
	instance->level_list = wl_create_level_list(i);
	wl_catalog_installed_levels(instance->level_list);
	instance->playlist = wl_create_playlist(instance->level_list->entry_count);
	if(!instance->playlist)
	{
		return false;
	}
	return true;
}

/* detect if we are playing the demo version */
bool wl_is_demo(void)
{
	if(!al_filename_exists("data/scripts/story_main/ending.script"))
	{
		return true;
	}
	return false;
}

static void _wl_handle_config_args(int argc, char * argv[])
{
	int i;

	for(i = 1; i < argc - 3; i++)
	{
		if(!strcmp(argv[i], "--set-config-value"))
		{
			al_set_config_value(t3f_config, argv[i + 1], argv[i + 2], argv[i + 3]);
		}
	}
}

static void _wl_handle_other_args(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, j;

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--reset-steam-stats"))
		{
			t3f_reset_steam_stats();
			for(j = 0; j < instance->achievements->entries; j++)
			{
				t3f_update_achievement_progress(instance->achievements, j, 0);
				instance->achievements->entry[j].store_state = 0;
			}
		}
	}
}

static void _wl_import_profile(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;

	for(i = 1; i < argc - 1; i++)
	{
		if(!strcmp(argv[i], "--import_profile"))
		{
			instance->import_profile_fn = argv[i + 1];
		}
	}
}

static void _wl_reset_profile(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[256];
	int i;

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--reset-profile"))
		{
			strcpy(buf, instance->profile->name);
			wl_destroy_profile(instance->profile);
			instance->profile = wl_create_profile();
			strcpy(instance->profile->name, buf);
		}
	}
}

static void _wl_unlock_profile(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--unlock") || !strcmp(argv[i], "--unlocked"))
		{
			instance->profile->flags |= WL_PROFILE_FLAG_UNLOCKED;
		}
	}
}

static void _wl_lock_profile(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;

	for(i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "--lock"))
		{
			instance->profile->flags &= ~WL_PROFILE_FLAG_UNLOCKED;
		}
	}
}

static void _wl_setup_urls(void)
{
	const char * val;

	val = al_get_config_value(t3f_config, "Config", "leaderboard_upload_url");
	if(!val)
	{
		al_set_config_value(t3f_config, "Config", "leaderboard_upload_url", "https://www.tcubedsoftware.com/scripts/leaderboards/update.php");
	}
	val = al_get_config_value(t3f_config, "Config", "leaderboard_download_url");
	if(!val)
	{
		al_set_config_value(t3f_config, "Config", "leaderboard_download_url", "https://www.tcubedsoftware.com/scripts/leaderboards/query.php");
	}
	val = al_get_config_value(t3f_config, "Config", "leaderboard_key_url");
	if(!val)
	{
		al_set_config_value(t3f_config, "Config", "leaderboard_key_url", "https://www.tcubedsoftware.com/scripts/leaderboards/get_user_key.php");
	}
	val = al_get_config_value(t3f_config, "Config", "leaderboard_username_url");
	if(!val)
	{
		al_set_config_value(t3f_config, "Config", "leaderboard_username_url", "https://www.tcubedsoftware.com/scripts/leaderboards/set_user_name.php");
	}
}

static void _wl_setup_controllers(WL_INSTANCE * instance)
{
	const char * val;

	val = al_get_config_value(t3f_config, "Config", "dead_zone");
	if(val)
	{
		instance->dead_zone = atof(val);
	}
}

bool wl_initialize(int argc, char * argv[], void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	const char * val;

	if(!t3f_initialize(T3F_APP_TITLE, WL_VIRTUAL_DISPLAY_WIDTH, WL_VIRTUAL_DISPLAY_HEIGHT, 60.0, wl_logic, wl_render, T3F_DEFAULT | T3F_FORCE_ASPECT | T3F_USE_FULLSCREEN | T3F_USE_FIXED_PIPELINE | T3F_FILL_SCREEN, data))
	{
		printf("Failed to initialize program!\n");
		return false;
	}
	t3f_constrain_view_aspect_ratio(t3f_default_view, 4.0 / 3.0, 16.0 / 9.0);

	#ifdef ALLEGRO_ANDROID
		t3net_setup(t3f_run_url, al_path_cstr(t3f_temp_path, '/'));
	#else
		t3net_setup(NULL, al_path_cstr(t3f_temp_path, '/'));
	#endif

	_wl_handle_config_args(argc, argv);
	_wl_import_profile(argc, argv, data);

	if(!wl_set_up_achievements(instance))
	{
		return false;
	}
	t3f_load_achievements_data(instance->achievements, t3f_user_data, "Achievements");

	instance->steam_running = t3f_init_steam_integration(instance->achievements);

	_wl_handle_other_args(argc, argv, data);

	_wl_setup_urls();
	_wl_setup_controllers(instance);
	instance->demo = wl_is_demo();

	t3f_set_event_handler(wl_process_event);

	wl_load_preferences(data);
	wl_initialize_video_mode(data);
	srand(time(0));
	t3f_srand(&instance->rng_state, time(0));

	/* create input handler for reading gamepads */
	if(!t3f_initialize_input(0))
	{
		return false;
	}
	instance->input_handler = t3f_create_input_handler(T3F_INPUT_HANDLER_TYPE_GAMEPAD);
	if(!instance->input_handler)
	{
		return false;
	}
	if(al_get_num_joysticks() > 0)
	{
		t3f_map_input_for_xbox_controller(instance->input_handler, 0);
	}

	/* select input type (default to mouse if available) */
	wl_select_input(wl_get_default_input_type(), 0, data);

	if(!wl_load_data(data))
	{
		return false;
	}
	_wl_reset_profile(argc, argv, data);
	_wl_lock_profile(argc, argv, data);
	_wl_unlock_profile(argc, argv, data);
	if(instance->steam_running)
	{
		val = t3f_get_steam_user_display_name();
		if(val)
		{
			strcpy(instance->profile->name, val);
		}
		instance->selected_profile = 0;
	}

	instance->game_messages.messages = 0;
	if(!wl_title_load_data(data, false))
	{
		return false;
	}
	if(!wl_title_init(data))
	{
		return false;
	}
	if(instance->option[WL_OPTION_FIRST_RUN])
	{
		wl_create_options_menu("INITIAL SETUP", "DONE", data);
		wl_select_menu(WL_MENU_OPTIONS, data);
	}
	else if(instance->import_profile_fn)
	{
		wl_create_import_profile_menu(data);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	wl_play_music(instance->music_queue, "data/music/theme.ogg");
	wl_fade_to(instance->effects_data, WL_BASE_COLOR, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), 60);
	return true;
}
