#include "t3f/t3f.h"
#include "t3f/file_utils.h"
#include "instance.h"
#include "character.h"

static bool _is_character_file(const char * fn, bool isfolder)
{
  ALLEGRO_PATH * path;
  ALLEGRO_CONFIG * cp;
  const char * val;
  bool ret = false;

  if(!isfolder)
  {
    path = al_create_path(fn);
    if(path)
    {
      if(!strcasecmp(al_get_path_extension(path), ".ini"))
      {
        cp = al_load_config_file(fn);
        if(cp)
        {
          val = al_get_config_value(cp, "Character", "name");
          if(val)
          {
            ret = true;
          }
          al_destroy_config(cp);
        }
      }
      al_destroy_path(path);
    }
  }
  return ret;
}

static bool _add_character_file(const char * fn, bool isfolder, void * data)
{
  WL_INSTANCE * instance = (WL_INSTANCE *)data;

  if(_is_character_file(fn, isfolder))
  {
    if(instance->character_count < WL_MAX_CHARACTERS)
    {
      instance->character[instance->character_count] = wl_load_character(fn, instance->graphics_scale);
      if(instance->character[instance->character_count])
      {
        instance->character_count++;
      }
    }
  }
  return false;
}

bool wl_load_characters(void * data)
{
  return t3f_scan_files("data/characters", _add_character_file, true, data);
}

int wl_get_character_by_name(const char * name, void * data)
{
  WL_INSTANCE * instance = (WL_INSTANCE *)data;
  int i;

  for(i = 0; i < instance->character_count; i++)
  {
    if(!strcmp(instance->character[i]->name, name))
    {
      return i;
    }
  }
  return -1;
}
