#include "t3f/t3f.h"
#include "instance.h"

WL_EFFECTS_DATA * wl_create_effects_data(void)
{
	WL_EFFECTS_DATA * dp;

	dp = malloc(sizeof(WL_EFFECTS_DATA));
	if(!dp)
	{
		goto fail;
	}
	memset(dp, 0, sizeof(WL_EFFECTS_DATA));
	return dp;

	fail:
	{
		return NULL;
	}
}

void wl_destroy_effects_data(WL_EFFECTS_DATA * dp)
{
	free(dp);
}

ALLEGRO_COLOR wl_fade_color(ALLEGRO_COLOR c1, ALLEGRO_COLOR c2, float amount)
{
	float r1, g1, b1, a1;
	float r2, g2, b2, a2;
	
	al_unmap_rgba_f(c1, &r1, &g1, &b1, &a1);
	al_unmap_rgba_f(c2, &r2, &g2, &b2, &a2);
	return al_map_rgba_f(r1 + (r2 - r1) * amount, g1 + (g2 - g1) * amount, b1 + (b2 - b1) * amount, a1 + (a2 - a1) * amount);
}

ALLEGRO_COLOR wl_darken_color(ALLEGRO_COLOR c1, float amount)
{
	float r1, g1, b1, a1;
	
	al_unmap_rgba_f(c1, &r1, &g1, &b1, &a1);
	return al_map_rgba_f(r1 * amount, g1 * amount, b1 * amount, a1);
}

ALLEGRO_COLOR wl_alpha_color(ALLEGRO_COLOR c1, float alpha)
{
	float r1, g1, b1, a1;
	
	al_unmap_rgba_f(c1, &r1, &g1, &b1, &a1);
	return al_map_rgba_f(r1 * alpha, g1 * alpha, b1 * alpha, a1 * alpha);
}

void wl_fade_to(WL_EFFECTS_DATA * dp, ALLEGRO_COLOR start_color, ALLEGRO_COLOR end_color, int time)
{
	dp->fade_time = time;
	dp->fade_tick = time;
	dp->fade_start_color = start_color;
	dp->fade_end_color = end_color;
	dp->fading = true;
}

void wl_stop_fade(WL_EFFECTS_DATA * dp)
{
	dp->fade_tick = 0;
	dp->fade_current_color = al_map_rgba_f(0.0, 0.0, 0.0, 0.0);
}

void wl_process_effects(WL_EFFECTS_DATA * dp)
{
	dp->fade_last_tick = dp->fade_tick;
	if(dp->fade_tick > 0)
	{
		dp->fade_tick--;
		if(dp->fade_tick <= 0)
		{
			dp->fading = false;
		}
	}
	if(dp->fade_last_tick != dp->fade_tick)
	{
		dp->fade_current_color = wl_fade_color(dp->fade_start_color, dp->fade_end_color, 1.0 - (float)dp->fade_tick / (float)dp->fade_time);
	}
}

void wl_render_effects(WL_EFFECTS_DATA * dp)
{
	unsigned char r, a;

	al_unmap_rgba(dp->fade_current_color, &r, &r, &r, &a);
	if(a > 0)
	{
		al_draw_filled_rectangle(0.0, 0.0, WL_VIRTUAL_DISPLAY_WIDTH, WL_VIRTUAL_DISPLAY_HEIGHT, dp->fade_current_color);
	}
}
