#include "t3f/t3f.h"
#include "instance.h"
#include "input.h"

int wl_get_default_input_type(void)
{
	if(t3f_flags & T3F_USE_MOUSE)
	{
    return WL_INPUT_TYPE_MOUSE;
	}
	else if(t3f_flags & T3F_USE_TOUCH)
	{
		return WL_INPUT_TYPE_TOUCH;
	}
	else if(al_get_num_joysticks() > 0)
	{
		return WL_INPUT_TYPE_GAMEPAD;
	}
	else
	{
		return WL_INPUT_TYPE_KEYBOARD;
	}
  return -1;
}

void wl_select_input(int type, int device, void * data)
{
  WL_INSTANCE * instance = (WL_INSTANCE *)data;
  int min_joystick = 0;
  #ifdef ALLEGRO_ANDROID
    min_joystick = 1;
  #endif

  if(type != instance->input_type || device != instance->input_device)
  {
    switch(type)
    {
      case WL_INPUT_TYPE_MOUSE:
      case WL_INPUT_TYPE_TOUCH:
      case WL_INPUT_TYPE_KEYBOARD:
      {
        instance->input_type = type;
        instance->input_device = device;
        instance->input_device_handle = NULL;
        break;
      }
      case WL_INPUT_TYPE_GAMEPAD:
      {
        if(device >= min_joystick)
        {
          instance->input_type = type;
          instance->input_device = device;
          instance->input_device_handle = t3f_joystick[device];
          t3f_map_input_for_xbox_controller(instance->input_handler, instance->input_device);
          instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].dead_zone = instance->dead_zone;
          instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].dead_zone = instance->dead_zone;
          instance->input_handler->element[T3F_GAMEPAD_RIGHT_ANALOG_X].dead_zone = instance->dead_zone;
          instance->input_handler->element[T3F_GAMEPAD_RIGHT_ANALOG_Y].dead_zone = instance->dead_zone;
        }
        break;
      }
    }
  }
}

void wl_read_user_input(void * data)
{
  WL_INSTANCE * instance = (WL_INSTANCE *)data;

  memset(&instance->input_state, 0, sizeof(WL_INPUT_STATE));
  switch(instance->input_type)
  {
    case WL_INPUT_TYPE_KEYBOARD:
    {
      /* up */
      if(t3f_key_held(ALLEGRO_KEY_UP) || t3f_key_held(ALLEGRO_KEY_W))
      {
        instance->input_state.up.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_UP) || t3f_key_pressed(ALLEGRO_KEY_W))
      {
        instance->input_state.up.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_UP);
        t3f_use_key_press(ALLEGRO_KEY_W);
      }
      if(t3f_key_released(ALLEGRO_KEY_UP) || t3f_key_released(ALLEGRO_KEY_W))
      {
        instance->input_state.up.released = true;
        t3f_use_key_release(ALLEGRO_KEY_UP);
        t3f_use_key_release(ALLEGRO_KEY_W);
      }

      /* down */
      if(t3f_key_held(ALLEGRO_KEY_DOWN) || t3f_key_held(ALLEGRO_KEY_S))
      {
        instance->input_state.down.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_DOWN) || t3f_key_pressed(ALLEGRO_KEY_S))
      {
        instance->input_state.down.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_DOWN);
        t3f_use_key_press(ALLEGRO_KEY_S);
      }
      if(t3f_key_released(ALLEGRO_KEY_DOWN) || t3f_key_released(ALLEGRO_KEY_S))
      {
        instance->input_state.down.released = true;
        t3f_use_key_release(ALLEGRO_KEY_DOWN);
        t3f_use_key_release(ALLEGRO_KEY_S);
      }

      /* left */
      if(t3f_key_held(ALLEGRO_KEY_LEFT) || t3f_key_held(ALLEGRO_KEY_A))
      {
        instance->input_state.left.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_LEFT) || t3f_key_pressed(ALLEGRO_KEY_A))
      {
        instance->input_state.left.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_LEFT);
        t3f_use_key_press(ALLEGRO_KEY_A);
      }
      if(t3f_key_released(ALLEGRO_KEY_LEFT) || t3f_key_released(ALLEGRO_KEY_A))
      {
        instance->input_state.left.released = true;
        t3f_use_key_release(ALLEGRO_KEY_LEFT);
        t3f_use_key_release(ALLEGRO_KEY_A);
      }

      /* right */
      if(t3f_key_held(ALLEGRO_KEY_RIGHT) || t3f_key_held(ALLEGRO_KEY_D))
      {
        instance->input_state.right.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_RIGHT) || t3f_key_pressed(ALLEGRO_KEY_D))
      {
        instance->input_state.right.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_RIGHT);
        t3f_use_key_press(ALLEGRO_KEY_D);
      }
      if(t3f_key_released(ALLEGRO_KEY_RIGHT) || t3f_key_released(ALLEGRO_KEY_D))
      {
        instance->input_state.right.released = true;
        t3f_use_key_release(ALLEGRO_KEY_RIGHT);
        t3f_use_key_release(ALLEGRO_KEY_D);
      }

      /* action */
      if(t3f_key_held(ALLEGRO_KEY_SPACE) || t3f_key_held(ALLEGRO_KEY_ENTER))
      {
        instance->input_state.action.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_SPACE) || t3f_key_pressed(ALLEGRO_KEY_ENTER))
      {
        instance->input_state.action.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_SPACE);
        t3f_use_key_press(ALLEGRO_KEY_ENTER);
      }
      if(t3f_key_released(ALLEGRO_KEY_SPACE) || t3f_key_released(ALLEGRO_KEY_ENTER))
      {
        instance->input_state.action.released = true;
        t3f_use_key_release(ALLEGRO_KEY_SPACE);
        t3f_use_key_release(ALLEGRO_KEY_ENTER);
      }

      /* pause/cancel */
      if(t3f_key_held(ALLEGRO_KEY_ESCAPE) || t3f_key_held(ALLEGRO_KEY_BACK))
      {
        instance->input_state.pause.held = true;
        instance->input_state.cancel.held = true;
      }
      if(t3f_key_pressed(ALLEGRO_KEY_ESCAPE) || t3f_key_pressed(ALLEGRO_KEY_BACK))
      {
        instance->input_state.pause.pressed = true;
        instance->input_state.cancel.pressed = true;
        t3f_use_key_press(ALLEGRO_KEY_ESCAPE);
        t3f_use_key_press(ALLEGRO_KEY_BACK);
      }
      if(t3f_key_released(ALLEGRO_KEY_ESCAPE) || t3f_key_released(ALLEGRO_KEY_BACK))
      {
        instance->input_state.pause.released = true;
        instance->input_state.cancel.released = true;
        t3f_use_key_release(ALLEGRO_KEY_ESCAPE);
        t3f_use_key_release(ALLEGRO_KEY_BACK);
      }
      break;
    }
    case WL_INPUT_TYPE_GAMEPAD:
    {
      /* up */
      instance->input_state.up.held = instance->input_handler->element[T3F_GAMEPAD_DPAD_UP].held || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].held && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val < 0.0);
      instance->input_state.up.pressed = instance->input_handler->element[T3F_GAMEPAD_DPAD_UP].pressed || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].pressed && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val < 0.0);
      instance->input_state.up.released = instance->input_handler->element[T3F_GAMEPAD_DPAD_UP].released || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].released && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val < 0.0);

      /* down */
      instance->input_state.down.held = instance->input_handler->element[T3F_GAMEPAD_DPAD_DOWN].held || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].held && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val > 0.0);
      instance->input_state.down.pressed = instance->input_handler->element[T3F_GAMEPAD_DPAD_DOWN].pressed || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].pressed && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val > 0.0);
      instance->input_state.down.released = instance->input_handler->element[T3F_GAMEPAD_DPAD_DOWN].released || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].released && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_Y].val > 0.0);

      /* left */
      instance->input_state.left.held = instance->input_handler->element[T3F_GAMEPAD_DPAD_LEFT].held || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].held && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val < 0.0);
      instance->input_state.left.pressed = instance->input_handler->element[T3F_GAMEPAD_DPAD_LEFT].pressed || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].pressed && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val < 0.0);
      instance->input_state.left.released = instance->input_handler->element[T3F_GAMEPAD_DPAD_LEFT].released || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].released && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val < 0.0);

      /* right */
      instance->input_state.right.held = instance->input_handler->element[T3F_GAMEPAD_DPAD_RIGHT].held || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].held && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val > 0.0);
      instance->input_state.right.pressed = instance->input_handler->element[T3F_GAMEPAD_DPAD_RIGHT].pressed || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].pressed && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val > 0.0);
      instance->input_state.right.released = instance->input_handler->element[T3F_GAMEPAD_DPAD_RIGHT].released || (instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].released && instance->input_handler->element[T3F_GAMEPAD_LEFT_ANALOG_X].val > 0.0);

      /* action */
      instance->input_state.action.held = instance->input_handler->element[T3F_GAMEPAD_A].held || instance->input_handler->element[T3F_GAMEPAD_X].held || instance->input_handler->element[T3F_GAMEPAD_Y].held || instance->input_handler->element[T3F_GAMEPAD_L].held || instance->input_handler->element[T3F_GAMEPAD_R].held || instance->input_handler->element[T3F_GAMEPAD_LEFT_TRIGGER].held || instance->input_handler->element[T3F_GAMEPAD_RIGHT_TRIGGER].held;
      instance->input_state.action.pressed = instance->input_handler->element[T3F_GAMEPAD_A].pressed || instance->input_handler->element[T3F_GAMEPAD_X].pressed || instance->input_handler->element[T3F_GAMEPAD_Y].pressed || instance->input_handler->element[T3F_GAMEPAD_L].pressed || instance->input_handler->element[T3F_GAMEPAD_R].pressed || instance->input_handler->element[T3F_GAMEPAD_LEFT_TRIGGER].pressed || instance->input_handler->element[T3F_GAMEPAD_RIGHT_TRIGGER].pressed;
      instance->input_state.action.released = instance->input_handler->element[T3F_GAMEPAD_A].released || instance->input_handler->element[T3F_GAMEPAD_X].released || instance->input_handler->element[T3F_GAMEPAD_Y].released || instance->input_handler->element[T3F_GAMEPAD_L].released || instance->input_handler->element[T3F_GAMEPAD_R].released || instance->input_handler->element[T3F_GAMEPAD_LEFT_TRIGGER].released || instance->input_handler->element[T3F_GAMEPAD_RIGHT_TRIGGER].released;

      /* cancel */
      instance->input_state.cancel.held = instance->input_handler->element[T3F_GAMEPAD_B].held;
      instance->input_state.cancel.pressed = instance->input_handler->element[T3F_GAMEPAD_B].pressed;
      instance->input_state.cancel.released = instance->input_handler->element[T3F_GAMEPAD_B].released;

      /* pause */
      instance->input_state.pause.held = instance->input_handler->element[T3F_GAMEPAD_SELECT].held || instance->input_handler->element[T3F_GAMEPAD_START].held;
      instance->input_state.pause.pressed = instance->input_handler->element[T3F_GAMEPAD_SELECT].pressed || instance->input_handler->element[T3F_GAMEPAD_START].pressed;
      instance->input_state.pause.released = instance->input_handler->element[T3F_GAMEPAD_SELECT].released || instance->input_handler->element[T3F_GAMEPAD_START].released;

      break;
    }
    case WL_INPUT_TYPE_MOUSE:
    {
      instance->input_state.pointer.x = t3f_get_mouse_x();
      instance->input_state.pointer.y = t3f_get_mouse_y();
      if(t3f_mouse_button_held(0) || t3f_mouse_button_held(1))
      {
        instance->input_state.action.held = true;
      }
      if(t3f_mouse_button_pressed(0) || t3f_mouse_button_pressed(1))
      {
        instance->input_state.action.pressed = true;
        t3f_use_mouse_button_press(0);
        t3f_use_mouse_button_press(1);
      }
      if(t3f_mouse_button_released(0) || t3f_mouse_button_released(1))
      {
        instance->input_state.action.released = true;
        t3f_use_mouse_button_release(0);
        t3f_use_mouse_button_release(1);
      }
      break;
    }
    case WL_INPUT_TYPE_TOUCH:
    {
      if(t3f_touch_active(1))
      {
        instance->input_state.pointer.x = t3f_get_touch_x(1);
        instance->input_state.pointer.y = t3f_get_touch_y(1);
        instance->input_state.action.held = true;
      }
      if(t3f_touch_released(1))
      {
        instance->input_state.pointer.x = t3f_get_touch_x(1);
        instance->input_state.pointer.y = t3f_get_touch_y(1);
        instance->input_state.action.released = true;
        t3f_use_touch_release(1);
      }
      if(t3f_touch_pressed(1))
      {
        instance->input_state.pointer.x = t3f_get_touch_x(1);
        instance->input_state.pointer.y = t3f_get_touch_y(1);
        instance->input_state.action.pressed = true;
        t3f_use_touch_press(1);
      }
      break;
    }
  }
}
