#include "t3f/animation.h"
#include "defines.h"
#include "particle.h"

void wl_initialize_particle(WL_PARTICLE * pp, int type, T3F_ANIMATION * ap, ALLEGRO_COLOR color, float x, float y, float z, float s, int life, int flags)
{
	pp->type = type;
	pp->animation = ap;
	pp->color = color;
	pp->x = x;
	pp->y = y;
	pp->z = z;
	pp->s = s;
	pp->start_life = life;
	pp->life = life;
	pp->flags = flags;
	pp->active = true;
}

void wl_process_particle(WL_PARTICLE * pp)
{
	if(pp->active)
	{
		switch(pp->type)
		{
			case WL_PARTICLE_TYPE_FLASH:
			case WL_PARTICLE_TYPE_MFLASH:
			{
				pp->life--;
				if(pp->life <= 0)
				{
					pp->active = false;
				}
				break;
			}
			case WL_PARTICLE_TYPE_GLASS:
			{
				pp->x += pp->vx;
				pp->y += pp->vy;
				pp->vy += 0.5;
				pp->z += pp->vz;
				pp->life--;
				if(pp->life <= 0)
				{
					pp->active = false;
				}
				break;
			}
			case WL_PARTICLE_TYPE_SHARD:
			{
				pp->x += pp->vx;
				pp->y += pp->vy;
				pp->z += pp->vz;
				pp->vz += 0.5;
				if(pp->vz > WL_TERMINAL_VELOCITY)
				{
					pp->vz = WL_TERMINAL_VELOCITY;
				}
				pp->life--;
				if(pp->life <= 0 || pp->z > 0.0)
				{
					pp->active = false;
				}
				break;
			}
			case WL_PARTICLE_TYPE_ICE:
			{
				pp->x += pp->vx;
				pp->y += pp->vy;
				pp->vy += 0.5;
				if(pp->vy > WL_TERMINAL_VELOCITY)
				{
					pp->vy = WL_TERMINAL_VELOCITY;
				}
				pp->z += pp->vz;
				pp->life--;
				if(pp->y > WL_VIRTUAL_DISPLAY_HEIGHT)
				{
					pp->active = false;
				}
				break;
			}
			case WL_PARTICLE_TYPE_STEAM:
			{
				pp->x += pp->vx;
				pp->y += pp->vy;
				pp->z += pp->vz;
				pp->life--;
				if(pp->life <= 0)
				{
					pp->active = false;
				}
				break;
			}
		}
	}
}

void wl_render_particle(WL_PARTICLE * pp)
{
	ALLEGRO_COLOR color;
	float r, g, b, a, aa, s;

	if(pp->active)
	{
		switch(pp->type)
		{
			case WL_PARTICLE_TYPE_FLASH:
			case WL_PARTICLE_TYPE_MFLASH:
			{
				if(pp->life < 10)
				{
					al_unmap_rgb_f(pp->color, &r, &g, &b);
					a = ((float)pp->life / 10.0) / 2.0;
					color = al_map_rgba_f(r * a, g * a, b * a, a * a);
					s = (float)pp->s / pp->animation->data->frame[0]->width;
					t3f_draw_scaled_animation(pp->animation, color, 0, pp->x, pp->y, 0.0, s, 0);
				}
				break;
			}
			case WL_PARTICLE_TYPE_GLASS:
			{
				al_unmap_rgba_f(pp->color, &r, &g, &b, &a);
				aa = ((float)pp->life / (float)pp->start_life) * a;
				color = al_map_rgba_f(r * aa, g * aa, b * aa, aa);
				t3f_draw_animation(pp->animation, color, 0, pp->x, pp->y, pp->z, 0);
				break;
			}
			case WL_PARTICLE_TYPE_SHARD:
			{
				al_unmap_rgb_f(pp->color, &r, &g, &b);
				a = (float)pp->life / (float)pp->start_life;
				color = al_map_rgba_f(r * a, g * a, b * a, a * a);
				t3f_draw_animation(pp->animation, color, 0, WL_BOARD_X + pp->x, WL_BOARD_Y + pp->y + pp->z, 0, 0);
				break;
			}
			case WL_PARTICLE_TYPE_ICE:
			{
				al_unmap_rgb_f(pp->color, &r, &g, &b);
				a = (float)pp->life / (float)pp->start_life;
				color = al_map_rgba_f(r * a, g * a, b * a, a * a);
				t3f_draw_animation(pp->animation, color, 0, pp->x, pp->y, pp->z, 0);
				break;
			}
			case WL_PARTICLE_TYPE_STEAM:
			{
				al_unmap_rgb_f(pp->color, &r, &g, &b);
				a = (float)pp->life / (float)pp->start_life;
				color = al_map_rgba_f(r * a, g * a, b * a, a * a);
				t3f_draw_animation(pp->animation, color, 0, pp->x, pp->y, pp->z, 0);
				break;
			}
		}
	}
}

WL_PARTICLE_LIST * wl_create_particle_list(int max)
{
	WL_PARTICLE_LIST * lp;

	lp = malloc(sizeof(WL_PARTICLE_LIST));
	if(!lp)
	{
		goto fail;
	}
	memset(lp, 0, sizeof(WL_PARTICLE_LIST));
	lp->particle = malloc(sizeof(WL_PARTICLE) * max);
	if(!lp->particle)
	{
		goto fail;
	}
	lp->particles = max;
	memset(lp->particle, 0, sizeof(WL_PARTICLE) * max);
	return lp;

	fail:
	{
		wl_destroy_particle_list(lp);
		return NULL;
	}
}

void wl_destroy_particle_list(WL_PARTICLE_LIST * lp)
{
	if(lp)
	{
		if(lp->particle)
		{
			free(lp->particle);
		}
		free(lp);
	}
}

void wl_reset_particle_list(WL_PARTICLE_LIST * lp)
{
	int i;

	for(i = 0; i < lp->particles; i++)
	{
		lp->particle[i].active = false;
	}
	lp->current_particle = 0;
}

int wl_add_particle(WL_PARTICLE_LIST * lp, int type, T3F_ANIMATION * ap, ALLEGRO_COLOR color, float x, float y, float z, float s, int life, int flags)
{
	int r = lp->current_particle;

	wl_initialize_particle(&lp->particle[lp->current_particle], type, ap, color, x, y, z, s, life, flags);
	lp->current_particle++;
	if(lp->current_particle >= lp->particles)
	{
		lp->current_particle = 0;
	}
	return r;
}

void wl_process_particle_list(WL_PARTICLE_LIST * lp)
{
	int i;

	for(i = 0; i < lp->particles; i++)
	{
		wl_process_particle(&lp->particle[i]);
	}
}

void wl_render_particle_list(WL_PARTICLE_LIST * lp, int type)
{
	int i;

	for(i = 0; i < lp->particles; i++)
	{
		if(type < 0 || lp->particle[i].type == type)
		{
			wl_render_particle(&lp->particle[i]);
		}
	}
}
