#include "t3f/t3f.h"
#include "data.h"
#include "text_entry.h"
#include "ui/menu_dialog.h"

WL_TEXT_ENTRY_DATA * wl_create_text_entry_data(void)
{
	WL_TEXT_ENTRY_DATA * text_entry;

	text_entry = malloc(sizeof(WL_TEXT_ENTRY_DATA));
	if(!text_entry)
	{
		goto fail;
	}
	memset(text_entry, 0, sizeof(WL_TEXT_ENTRY_DATA));
	text_entry->entered_text_ustr = al_ustr_new("\\_");
	if(!text_entry->entered_text_ustr)
	{
		goto fail;
	}

	return text_entry;

	fail:
	{
		wl_destroy_text_entry_data(text_entry);
		return NULL;
	}
}

void wl_destroy_text_entry_data(WL_TEXT_ENTRY_DATA * text_entry)
{
	if(text_entry)
	{
		if(text_entry->entered_text_ustr)
		{
			al_ustr_free(text_entry->entered_text_ustr);
		}
		free(text_entry);
	}
}

bool wl_is_string_empty(WL_TEXT_ENTRY_DATA * text_entry)
{
	int i;

	if(al_ustr_length(text_entry->entered_text_ustr) <= 2)
	{
		return true;
	}
	for(i = 0; i < (int)al_ustr_length(text_entry->entered_text_ustr); i++)
	{
		if(al_ustr_get(text_entry->entered_text_ustr, i) != ' ')
		{
			return false;
		}
	}
	return true;
}

void wl_enter_text(WL_TEXT_ENTRY_DATA * text_entry, void * data)
{
	if(text_entry->entering_text)
	{
		char key = t3f_get_char(0);
		switch(key)
		{
			case '\r':
			{
				wl_end_text_entry(text_entry);
				wl_menu_proc_play_new_ok(data, 0, NULL);
				break;
			}
			case '\b':
			case 127:
			{
				if(text_entry->entering_text_pos > 0)
				{
					al_ustr_remove_chr(text_entry->entered_text_ustr, text_entry->entering_text_pos - 1);
					text_entry->entering_text_pos--;
					al_ustr_to_buffer(text_entry->entered_text_ustr, text_entry->entered_text, WL_TEXT_ENTRY_MAX);
				}
				break;
			}

			/* ignore these keys */
			case '\t':
			case '\0':
			{
				break;
			}

			/* type these keys */
			default:
			{
				if(text_entry->entering_text_pos < 18)
				{
					al_ustr_insert_chr(text_entry->entered_text_ustr, text_entry->entering_text_pos, key);
					text_entry->entering_text_pos++;
					al_ustr_to_buffer(text_entry->entered_text_ustr, text_entry->entered_text, 256);
				}
				break;
			}
		}
	}
}

void wl_begin_text_entry(WL_TEXT_ENTRY_DATA * text_entry, const char * initial)
{
	if(!text_entry->entering_text)
	{
		t3f_clear_chars();
		al_ustr_assign_cstr(text_entry->entered_text_ustr, initial);
		al_ustr_append_cstr(text_entry->entered_text_ustr, "\\_");
		strcpy(text_entry->entered_text, al_cstr(text_entry->entered_text_ustr));
		text_entry->entering_text_pos = strlen(initial);
		text_entry->entering_text = 1;
	}
}

void wl_end_text_entry(WL_TEXT_ENTRY_DATA * text_entry)
{
	if(text_entry->entering_text)
	{
		text_entry->entered_text[text_entry->entering_text_pos] = 0;
		text_entry->entering_text = 0;
	}
}
