#include "t3f/t3f.h"
#include "tile.h"

WL_TILE * wl_load_tile(const char * fn, int flags)
{
  WL_TILE * tp = NULL;
  int i, c = 0, w = 0, h = 0;
  int fc;
  int atlas_flags[WL_TILE_MAX_BITMAPS] = {0};

  tp = malloc(sizeof(WL_TILE));
  if(!tp)
  {
    goto fail;
  }
  memset(tp, 0, sizeof(WL_TILE));

  atlas_flags[WL_TILE_STANDALONE] = T3F_ATLAS_SPRITE;
  atlas_flags[WL_TILE_H_LEFT] = T3F_ATLAS_TILE_CONNECT_RIGHT;
  atlas_flags[WL_TILE_H_CENTER] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_RIGHT;
  atlas_flags[WL_TILE_H_RIGHT] = T3F_ATLAS_TILE_CONNECT_LEFT;
  atlas_flags[WL_TILE_V_TOP] = T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_V_CENTER] = T3F_ATLAS_TILE_CONNECT_TOP | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_V_BOTTOM] = T3F_ATLAS_TILE_CONNECT_TOP;
  atlas_flags[WL_TILE_TOP_LEFT] = T3F_ATLAS_TILE_CONNECT_RIGHT | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_TOP] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_RIGHT | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_TOP_RIGHT] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_LEFT] = T3F_ATLAS_TILE_CONNECT_RIGHT | T3F_ATLAS_TILE_CONNECT_TOP | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_CENTER] = T3F_ATLAS_TILE;
  atlas_flags[WL_TILE_RIGHT] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_TOP | T3F_ATLAS_TILE_CONNECT_BOTTOM;
  atlas_flags[WL_TILE_BOTTOM_LEFT] = T3F_ATLAS_TILE_CONNECT_RIGHT | T3F_ATLAS_TILE_CONNECT_TOP;
  atlas_flags[WL_TILE_BOTTOM] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_RIGHT | T3F_ATLAS_TILE_CONNECT_TOP;
  atlas_flags[WL_TILE_BOTTOM_RIGHT] = T3F_ATLAS_TILE_CONNECT_LEFT | T3F_ATLAS_TILE_CONNECT_TOP;
  if(flags & WL_TILE_CONNECT_NONE)
  {
    tp->bitmap[WL_TILE_STANDALONE] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PADDED, false);
  }
  if(flags & WL_TILE_CONNECT_H)
  {
    tp->bitmap[WL_TILE_H_LEFT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_TOP | T3F_BITMAP_FLAG_PAD_BOTTOM, false);
    tp->bitmap[WL_TILE_H_CENTER] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_TOP | T3F_BITMAP_FLAG_PAD_BOTTOM, false);
    tp->bitmap[WL_TILE_H_RIGHT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_RIGHT | T3F_BITMAP_FLAG_PAD_TOP | T3F_BITMAP_FLAG_PAD_BOTTOM, false);
  }
  if(flags & WL_TILE_CONNECT_V)
  {
    tp->bitmap[WL_TILE_V_TOP] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_TOP | T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_RIGHT, false);
    tp->bitmap[WL_TILE_V_CENTER] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_RIGHT, false);
    tp->bitmap[WL_TILE_V_BOTTOM] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_BOTTOM | T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_RIGHT, false);
  }
  if(flags & WL_TILE_CONNECT_VH)
  {
    tp->bitmap[WL_TILE_TOP_LEFT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_TOP, false);
    tp->bitmap[WL_TILE_TOP] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_TOP, false);
    tp->bitmap[WL_TILE_TOP_RIGHT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_RIGHT | T3F_BITMAP_FLAG_PAD_TOP, false);
    tp->bitmap[WL_TILE_LEFT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_LEFT, false);
    tp->bitmap[WL_TILE_CENTER] = t3f_load_bitmap(fn, 0, false);
    tp->bitmap[WL_TILE_RIGHT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_RIGHT, false);
    tp->bitmap[WL_TILE_BOTTOM_LEFT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_LEFT | T3F_BITMAP_FLAG_PAD_BOTTOM, false);
    tp->bitmap[WL_TILE_BOTTOM] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_BOTTOM, false);
    tp->bitmap[WL_TILE_BOTTOM_RIGHT] = t3f_load_bitmap(fn, T3F_BITMAP_FLAG_PAD_RIGHT | T3F_BITMAP_FLAG_PAD_BOTTOM, false);
  }
  for(i = 0; i < WL_TILE_MAX_BITMAPS; i++)
  {
    if(tp->bitmap[i])
    {
      tp->width = tp->bitmap[i]->target_width;
      tp->height = tp->bitmap[i]->target_height;
      w = al_get_bitmap_width(tp->bitmap[i]->bitmap);
      h = al_get_bitmap_height(tp->bitmap[i]->bitmap);
      c++;
    }
  }

  fc = sqrt(c) + 0.5;
  tp->atlas = t3f_create_atlas((w + 4) * fc, (h + 4) * fc);
  if(tp->atlas)
  {
    for(i = 0; i < WL_TILE_MAX_BITMAPS; i++)
    {
      if(tp->bitmap[i])
      {
        t3f_add_bitmap_to_atlas(tp->atlas, &tp->bitmap[i]->bitmap, atlas_flags[i]);
      }
    }
  }

  return tp;

  fail:
  {
    if(tp)
    {
      wl_destroy_tile(tp);
    }
    return NULL;
  }
}

void wl_destroy_tile(WL_TILE * tp)
{
  int i;

  if(tp)
  {
    for(i = 0; i < WL_TILE_MAX_BITMAPS; i++)
    {
      if(tp->bitmap[i])
      {
        t3f_destroy_bitmap(tp->bitmap[i]);
      }
      if(tp->atlas)
      {
        t3f_destroy_atlas(tp->atlas);
      }
    }
  }
}

bool wl_add_tile_to_atlas(T3F_ATLAS * atlas, WL_TILE * tile)
{
  int i;
  bool ret = true;

  for(i = 0; i < WL_TILE_MAX_BITMAPS; i++)
  {
    if(tile->bitmap[i])
    {
      if(!t3f_add_bitmap_to_atlas(atlas, &tile->bitmap[i]->bitmap, T3F_ATLAS_TILE))
      {
        ret = false;
      }
    }
  }
  return ret;
}

void wl_draw_single_tile(WL_TILE * tp, int tile, ALLEGRO_COLOR color, float x, float y, float z, int flags)
{
  t3f_draw_bitmap(tp->bitmap[tile], color, x, y, z, flags);
}

void wl_draw_tile(WL_TILE * tp, ALLEGRO_COLOR color, float x, float y, float z, int w, int h, int flags)
{
  int i, j;

  if(w == 1 && h == 1)
  {
    t3f_draw_bitmap(tp->bitmap[WL_TILE_STANDALONE], color, x, y, z, flags);
  }
  else if(h == 1 && w > 0)
  {
    t3f_draw_bitmap(tp->bitmap[WL_TILE_H_LEFT], color, x, y, z, flags);
    for(i = 1; i < w - 1; i++)
    {
      t3f_draw_bitmap(tp->bitmap[WL_TILE_H_CENTER], color, x + i * tp->width, y, z, flags);
    }
    t3f_draw_bitmap(tp->bitmap[WL_TILE_H_RIGHT], color, x + tp->width * (w - 1), y, z, flags);
  }
  else if(w == 1 && h > 0)
  {
    t3f_draw_bitmap(tp->bitmap[WL_TILE_V_TOP], color, x, y, z, flags);
    for(i = 1; i < h - 1; i++)
    {
      t3f_draw_bitmap(tp->bitmap[WL_TILE_V_CENTER], color, x, y + i * tp->height, z, flags);
    }
    t3f_draw_bitmap(tp->bitmap[WL_TILE_V_BOTTOM], color, x, y + tp->width * (w - 1), z, flags);
  }
  else if(w > 0 && h > 0)
  {
    /* top row */
    t3f_draw_bitmap(tp->bitmap[WL_TILE_TOP_LEFT], color, x, y, z, flags);
    for(i = 1; i < w - 1; i++)
    {
      t3f_draw_bitmap(tp->bitmap[WL_TILE_TOP], color, x + i * tp->width, y, z, flags);
    }
    t3f_draw_bitmap(tp->bitmap[WL_TILE_TOP_RIGHT], color, x + tp->width * (w - 1), y, z, flags);

    /* middle section */
    for(i = 1; i < h - 1; i++)
    {
      t3f_draw_bitmap(tp->bitmap[WL_TILE_LEFT], color, x, y + i * tp->height, z, flags);
      for(j = 1; j < w - 1; j++)
      {
        t3f_draw_bitmap(tp->bitmap[WL_TILE_CENTER], color, x + j * tp->width, y + i * tp->height, z, flags);
      }
      t3f_draw_bitmap(tp->bitmap[WL_TILE_RIGHT], color, x + tp->width * (w - 1), y +  + i * tp->height, z, flags);
    }

    /* bottom row */
    t3f_draw_bitmap(tp->bitmap[WL_TILE_BOTTOM_LEFT], color, x, y + tp->height * (h - 1), z, flags);
    for(i = 1; i < w - 1; i++)
    {
      t3f_draw_bitmap(tp->bitmap[WL_TILE_BOTTOM], color, x + i * tp->width, y + tp->height * (h - 1), z, flags);
    }
    t3f_draw_bitmap(tp->bitmap[WL_TILE_BOTTOM_RIGHT], color, x + tp->width * (w - 1), y + tp->height * (h - 1), z, flags);
  }
}

void wl_draw_scaled_tile(WL_TILE * fp, ALLEGRO_COLOR color, float x, float y, float z, float width, float height, int w, int h, int flags);
