#include "t3f/t3f.h"
#include "t3f/resource.h"
#include "instance.h"
#include "defines.h"
#include "data.h"

bool wl_video_mode_okay(int mode)
{
	ALLEGRO_MONITOR_INFO info;

	al_get_monitor_info(0, &info);
	switch(mode)
	{
		case WL_VIDEO_MODE_MEDIUM_WINDOW:
		{
			if(info.x2 - info.x1 > 960 && info.y2 - info.y1 > 720)
			{
				return true;
			}
			return false;
		}
		case WL_VIDEO_MODE_LARGE_WINDOW:
		{
			if(info.x2 - info.x1 > 1280 && info.y2 - info.y1 > 960)
			{
				return true;
			}
			return false;
		}
		default:
		{
			return true;
		}
	}
}

bool wl_set_video_mode(int w, int h, int flags, void * data)
{
	/* must recreate display if we get error code 2 */
	if(t3f_set_gfx_mode(w, h, flags) == 2)
	{
		t3f_unload_resources();
		al_unregister_event_source(t3f_queue, al_get_display_event_source(t3f_display));
		al_destroy_display(t3f_display);
		t3f_display = NULL;
		t3f_set_gfx_mode(WL_VIRTUAL_DISPLAY_WIDTH, WL_VIRTUAL_DISPLAY_HEIGHT, flags);
		al_register_event_source(t3f_queue, al_get_display_event_source(t3f_display));
		t3f_reload_resources();
		t3f_default_view->need_update = true;
		t3f_select_view(t3f_default_view);
	}
	return true;
}

void wl_initialize_video_mode(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float scale_chart[5] = {1.0, 1.5, 2.0, 3.0, 4.0};
	float scale_select[5] = {1.0, 2.0, 2.0, 3.0, 4.0};
	int i;

	for(i = 0; i < 5; i++)
	{
		if(640.0 * scale_chart[i] > al_get_display_width(t3f_display))
		{
			break;
		}
		if(480.0 * scale_chart[i] > al_get_display_height(t3f_display))
		{
			break;
		}
	}
	i--;
	if(i < 1)
	{
		i = 0;
	}
	instance->graphics_scale = scale_select[i];
}
