#include "t3f/t3f.h"
#include "modules/data.h"
#include "defines.h"
#include "instance.h"
#include "preferences.h"
#include "menu_options.h"

static void _set_video_mode(int i)
{
	ALLEGRO_MONITOR_INFO info;
	int w, h, mode_w, mode_h, mode_flags = 0;
	char val[64];

	al_stop_timer(t3f_timer);
	if(i == 0)
	{
		al_set_config_value(t3f_config, "T3F", "force_window", "false");
		t3f_set_gfx_mode(640, 480, T3F_USE_FULLSCREEN);
	}
	else
	{
		al_set_config_value(t3f_config, "T3F", "force_window", "true");
		#ifndef ALLEGRO_MACOSX
			al_set_config_value(t3f_config, "T3F", "force_resizable", "true");
		#endif
		al_set_config_value(t3f_config, "T3F", "force_fullscreen", "false");
		al_get_monitor_info(0, &info);
		w = info.x2 - info.x1;
		h = info.y2 - info.y1;
		for(i = 1; i < 7; i++)
		{
			mode_w = wl_video_mode[i * 2];
			mode_h = wl_video_mode[i * 2 + 1];
			if(mode_w >= w || mode_h >= h)
			{
				break;
			}
		}
		i--;
		#ifndef ALLEGRO_MACOSX
			mode_flags = T3F_RESIZABLE;
		#endif
		t3f_set_gfx_mode(wl_video_mode[i * 2], wl_video_mode[i * 2 + 1], mode_flags);
	}
	al_start_timer(t3f_timer);
}

int _wl_menu_proc_options_video_mode_toggle(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_VIDEO_MODE] = instance->option[WL_OPTION_VIDEO_MODE] > 0 ? 0 : 1;
	wl_create_options_menu(NULL, NULL, data);
	_set_video_mode(instance->option[WL_OPTION_VIDEO_MODE]);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);

	return 1;
}

int wl_menu_proc_options_controls_left(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_CONTROLS]--;
	if(instance->option[WL_OPTION_CONTROLS] < 0)
	{
		instance->option[WL_OPTION_CONTROLS] = 2;
	}
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_controls_right(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_CONTROLS]++;
	if(instance->option[WL_OPTION_CONTROLS] > 2)
	{
		instance->option[WL_OPTION_CONTROLS] = 0;
	}
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_music_left(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_MUSIC_VOLUME] -= 5;
	if(instance->option[WL_OPTION_MUSIC_VOLUME] < 0)
	{
		instance->option[WL_OPTION_MUSIC_VOLUME] = 0;
	}
	wl_set_music_base_volume(instance->music_queue, (float)instance->option[WL_OPTION_MUSIC_VOLUME] / 100.0);
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_music_right(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_MUSIC_VOLUME] += 5;
	if(instance->option[WL_OPTION_MUSIC_VOLUME] > 200)
	{
		instance->option[WL_OPTION_MUSIC_VOLUME] = 200;
	}
	wl_set_music_base_volume(instance->music_queue, (float)instance->option[WL_OPTION_MUSIC_VOLUME] / 100.0);
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_sound_left(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_SOUND_VOLUME] -= 5;
	if(instance->option[WL_OPTION_SOUND_VOLUME] < 0)
	{
		instance->option[WL_OPTION_SOUND_VOLUME] = 0;
	}
	t3f_set_sound_volume((float)instance->option[WL_OPTION_SOUND_VOLUME] / 100.0);
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_sound_right(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_SOUND_VOLUME] += 5;
	if(instance->option[WL_OPTION_SOUND_VOLUME] > 200)
	{
		instance->option[WL_OPTION_SOUND_VOLUME] = 200;
	}
	t3f_set_sound_volume((float)instance->option[WL_OPTION_SOUND_VOLUME] / 100.0);
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_internet_left(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_UPLOAD_SCORES]--;
	if(instance->option[WL_OPTION_UPLOAD_SCORES] < 0)
	{
		instance->option[WL_OPTION_UPLOAD_SCORES] = 1;
	}
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_internet_right(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int last_item = instance->menu[instance->current_menu]->hover_element;

	instance->option[WL_OPTION_UPLOAD_SCORES]++;
	if(instance->option[WL_OPTION_UPLOAD_SCORES] > 1)
	{
		instance->option[WL_OPTION_UPLOAD_SCORES] = 0;
	}
	wl_create_options_menu(NULL, NULL, data);
	instance->menu[instance->current_menu]->hover_element = last_item;
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_options_back(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->option[WL_OPTION_FIRST_RUN] = 0;
	wl_save_preferences(data);
	wl_select_menu(WL_MENU_MAIN, data);
	return 1;
}

bool wl_create_options_menu(char * title, char * done_text, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char volume_text[128] = {0};
	char * internet_text[2] = {"NO", "YES"};
	float oy = 0.0;
	int aflags = 0;

	if(!title)
	{
		title = instance->options_title;
	}
	else
	{
		instance->options_title = title;
	}
	if(!done_text)
	{
		done_text = instance->options_done_text;
	}
	else
	{
		instance->options_done_text = done_text;
	}
	if(instance->menu[WL_MENU_OPTIONS])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_OPTIONS]);
	}
	#ifdef ALLEGRO_ANDROID
		aflags = 0;
	#endif
	instance->menu[WL_MENU_OPTIONS] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_OPTIONS])
	{
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, title, (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
		oy += 48;
		#ifndef ALLEGRO_ANDROID
			t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, "VIDEO MODE", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
			oy += 32;
			if(instance->option[WL_OPTION_VIDEO_MODE] == 0)
			{
				sprintf(volume_text, "FULL SCREEN");
			}
			else
			{
				sprintf(volume_text, "WINDOW");
			}
			t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, volume_text, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			t3f_describe_last_gui_element(instance->menu[WL_MENU_OPTIONS], "Current Video Mode");
			t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], _wl_menu_proc_options_video_mode_toggle, "<", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) - 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
			t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], _wl_menu_proc_options_video_mode_toggle, ">", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) + 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
			oy += 32;
		#endif
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, "MUSIC VOLUME", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
		oy += 32;
		sprintf(volume_text, "%d%%", instance->option[WL_OPTION_MUSIC_VOLUME]);
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, volume_text, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_OPTIONS], "Volume of Music Playback");
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_music_left, "<", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) - 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_music_right, ">", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) + 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		oy += 32;
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, "SOUND VOLUME", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
		oy += 32;
		sprintf(volume_text, "%d%%", instance->option[WL_OPTION_SOUND_VOLUME]);
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, volume_text, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_OPTIONS], "Volume of Sound Effects Playback");
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_sound_left, "<", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) - 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_sound_right, ">", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) + 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		oy += 32;
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, "UPLOAD SCORES", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
		oy += 32;
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], NULL, internet_text[instance->option[WL_OPTION_UPLOAD_SCORES]], (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_OPTIONS], "Score Upload Setting");
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_internet_left, "<", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) - 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_internet_right, ">", (void **)&instance->font[WL_FONT_16], (WL_VIRTUAL_DISPLAY_WIDTH / 2) + 180, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY | aflags);
		oy += 40;
		t3f_add_gui_text_element(instance->menu[WL_MENU_OPTIONS], wl_menu_proc_options_back, done_text, (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, oy, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_center_gui(instance->menu[WL_MENU_OPTIONS], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
  return true;
}
