#include "t3f/t3f.h"
#include "instance.h"
#include "gameplay/game_intro.h"
#include "ui/menu_dialog.h"

int wl_menu_proc_overlay_menu(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->overlay_pause_state = instance->state;
	if(instance->game.mode >= WL_GAME_MODE_STORY && instance->state != WL_STATE_GAME_TUTORIAL)
	{
		al_stop_timer(t3f_timer);
		t3f_render(true);
		instance->cinema = wl_load_cinema_script(instance->level_list->story[instance->game.player[0].story].pause_fn, instance->graphics_scale);
		if(instance->cinema)
		{
			wl_set_cinema_view_callbacks(instance->cinema, wl_game_paused_view_logic, wl_game_render);
			wl_show_pause_menu(data);
		}
		else
		{
			wl_show_pause_menu(data);
		}
		al_start_timer(t3f_timer);
		wl_block_cinema_skip();
	}
	else if(instance->state == WL_STATE_MUSIC)
	{
		wl_select_menu(WL_MENU_SCENE_PAUSED, data);
		instance->overlay_pause_state = instance->state;
		instance->state = WL_STATE_GAME_PAUSED;
		t3f_pause_music();
		wl_set_music_fade_speed(instance->music_queue, 0.0);
	}
	else
	{
		wl_show_pause_menu(data);
	}
	instance->menu[WL_MENU_GAME_OVERLAY]->hover_element = -1;
	instance->menu[WL_MENU_MUSIC_OVERLAY]->hover_element = -1;
	return 1;
}

int wl_menu_proc_overlay_fast_forward(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.fast_forward = true;
	return 1;
}

int wl_menu_proc_overlay_next_track(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->playlist_pos < instance->playlist->count - 1)
	{
		wl_clear_music_queue(instance->music_queue);
		instance->playlist_pos++;
		instance->game.player[0].start_level = instance->playlist->entry[instance->playlist_pos];
		wl_game_initialize(data, 0);
	}
	else
	{
		wl_finish_current_song(instance->music_queue, 0.0);
		instance->game.music_done = true;
	}
	instance->game.fast_forward = false;
	return 1;
}

int wl_menu_proc_overlay_previous_track(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	wl_clear_music_queue(instance->music_queue);
	if(instance->playlist_pos > 0)
	{
		instance->playlist_pos--;
	}
	instance->game.player[0].start_level = instance->playlist->entry[instance->playlist_pos];
	wl_game_initialize(data, 0);
	instance->game.fast_forward = false;
	return 1;
}

T3F_GUI * wl_create_game_overlay_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
  T3F_GUI * menu;

	menu = t3f_create_gui(0, 0);
	if(!menu)
	{
		return NULL;
	}
	t3f_add_gui_text_element(menu, wl_menu_proc_overlay_menu, "\\|", (void **)&instance->font[WL_FONT_24], WL_LIFE_BAR_X, WL_LEVEL_METER_Y, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 0);

  return menu;
}

T3F_GUI * wl_create_music_overlay_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
  T3F_GUI * menu;

	menu = t3f_create_gui(0, 0);
	if(!menu)
	{
		return NULL;
	}
	t3f_add_gui_text_element(menu, wl_menu_proc_overlay_previous_track, "\\-", (void **)&instance->font[WL_FONT_24], t3f_default_view->right - 16 - 32 * 4, t3f_default_view->bottom - 16 - 32, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 0);
	t3f_add_gui_text_element(menu, wl_menu_proc_overlay_menu, "\\|", (void **)&instance->font[WL_FONT_24], t3f_default_view->right - 16 - 32 * 3, t3f_default_view->bottom - 16 - 32, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 0);
	t3f_add_gui_text_element(menu, wl_menu_proc_overlay_fast_forward, "\\>", (void **)&instance->font[WL_FONT_24], t3f_default_view->right - 16 - 32 * 2, t3f_default_view->bottom - 16 - 32, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_ON_TOUCH);
	t3f_add_gui_text_element(menu, wl_menu_proc_overlay_next_track, "\\+", (void **)&instance->font[WL_FONT_24], t3f_default_view->right - 16 - 32 * 1, t3f_default_view->bottom - 16 - 32, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 0);

  return menu;
}
