#include "t3f/t3f.h"
#include "modules/character_list.h"
#include "instance.h"
#include "menu_play.h"

int wl_menu_proc_game_players_back(void * data, int i, void * p)
{
	wl_select_menu(WL_MENU_GAME, data);
	return 1;
}

int wl_menu_proc_play_cancel(void * data, int i, void * p)
{
	wl_select_menu(WL_MENU_GAME, data);
	return 1;
}

static void _wl_menu_start_game(WL_INSTANCE * instance, int level)
{
	instance->game.player[0].start_level = level;
	wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 60);
	wl_finish_current_song(instance->music_queue, 1.0 / 60.0);
	instance->state = WL_STATE_TO_GAME;
}

static int _wl_menu_proc_play_story_new(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_wl_menu_start_game(instance, -1);

	return 1;
}

static int _wl_menu_proc_play_story_continue(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_wl_menu_start_game(instance, instance->level_list->story[instance->game.player[0].story].start_level + instance->profile->progress[instance->game.player[0].story] / WL_LEVEL_UP);

	return 1;
}

int wl_menu_proc_play_select_level(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	T3F_GUI * pp = (T3F_GUI *)p;
	int level_index;

	level_index = instance->level_list->unlocked_index[pp->hover_element - 2 + instance->level_page * 9];
	al_set_config_value(t3f_user_data, "Preferences", "Level", instance->level_list->entry[level_index].name);
	wl_create_play_mode_other_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);
	return 1;
}

static int _wl_menu_proc_play_story_select_level_pick(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	T3F_GUI * pp = (T3F_GUI *)p;

	_wl_menu_start_game(instance, instance->level_list->story[instance->game.player[0].story].start_level + pp->hover_element - 1);

	return 1;
}

int wl_menu_proc_play_select_level_cancel(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->game.mode == WL_GAME_MODE_MUSIC)
	{
		wl_create_play_select_music_menu(data);
	}
	else
	{
		wl_create_play_story_select_menu(data, instance->game.player[0].story);
	}
	wl_select_menu(WL_MENU_DYNAMIC, data);
	return 1;
}

static int _wl_menu_proc_play_story_select_level(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int j;
	int pos = 0;
	int start, end;

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "SELECT LEVEL", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC);
		pos += 48;
		start = instance->level_list->story[instance->game.player[0].story].start_level;
		end = start + instance->profile->progress[instance->game.player[0].story] / WL_LEVEL_UP;
		if(instance->profile->flags & WL_PROFILE_FLAG_UNLOCKED)
		{
			end = instance->level_list->story[instance->game.player[0].story].end_level;
		}
		if(end > instance->level_list->story[instance->game.player[0].story].end_level)
		{
			end = instance->level_list->story[instance->game.player[0].story].end_level;
		}
		for(j = start; j <= end; j++)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_select_level_pick, instance->level_list->entry[j].name, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		pos += 8;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_level_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	return 1;
}

static int _wl_menu_proc_play_story_intro(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_wl_menu_start_game(instance, -1);

	return 1;
}

static int _wl_menu_proc_play_story_ending(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	_wl_menu_start_game(instance, instance->level_list->story[instance->game.player[0].story].end_level + 1);

	return 1;
}

static int _wl_menu_proc_play_story_cancel(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(instance->game.player[0].story == 0)
	{
		wl_create_play_mode_menu(data);
	}
	else
	{
		wl_create_play_extra_stories_menu(data);
	}
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

void wl_create_play_story_select_menu(void * data, int story)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int pos = 0;

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{

		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, instance->level_list->story[story].name, (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC);
		pos += 48;
		if(instance->profile->progress[story] <= 0 && !(instance->profile->flags & WL_PROFILE_FLAG_UNLOCKED))
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_new, "PLAY", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		else
		{
			if(instance->profile->progress[story] / WL_LEVEL_UP <= instance->level_list->story[story].end_level - instance->level_list->story[story].start_level)
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_continue, "CONTINUE", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
				pos += 32;
			}
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_select_level, "SELECT LEVEL", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_intro, "WATCH INTRO", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		if(!instance->demo)
		{
			if((instance->profile->flags & WL_PROFILE_FLAG_UNLOCKED) || instance->profile->progress[story] / WL_LEVEL_UP > instance->level_list->story[story].end_level - instance->level_list->story[story].start_level)
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_ending, "WATCH ENDING", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
				pos += 32;
			}
		}
		pos += 8;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_story_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
}

int wl_menu_proc_play_select_level_previous_page(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->level_page--;
	wl_create_play_select_menu(data);
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_play_select_level_next_page(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->level_page++;
	wl_create_play_select_menu(data);
	wl_reset_particle_list(instance->particle_list[0]);
	return 1;
}

int wl_menu_proc_play_select_random(void * data, int i, void * p)
{
	al_remove_config_key(t3f_user_data, "Preferences", "Level");
	wl_create_play_mode_other_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

int wl_menu_proc_play_select_cancel(void * data, int i, void * p)
{
	wl_create_play_mode_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

static int _wl_get_level_index_from_name(WL_LEVEL_LIST * lp, const char * name, T3F_RNG_STATE * rng)
{
	int i;

	if(name)
	{
		for(i = 0; i < lp->unlocked_count; i++)
		{
			if(!strcmp(name, lp->entry[lp->unlocked_index[i]].name))
			{
				return lp->unlocked_index[i];
			}
		}
	}
	return lp->unlocked_index[t3f_rand(rng) % lp->unlocked_count];
}

static void _wl_start_game_other(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	const char * val;

	val = al_get_config_value(t3f_user_data, "Preferences", "Level");
	instance->game.player[0].start_level = _wl_get_level_index_from_name(instance->level_list, val, &instance->rng_state);
	wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 60);
	wl_finish_current_song(instance->music_queue, 1.0 / 60.0);
	instance->state = WL_STATE_TO_GAME;
}

void wl_create_play_select_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, max;
	int pos = 0;
	int start_entry;

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{
		start_entry = instance->level_page * 9;
		max = start_entry + 9;
		if(max > instance->level_list->unlocked_count)
		{
			max = instance->level_list->unlocked_count;
		}

		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "SELECT LEVEL", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC);
		pos += 48;
		if(instance->level_page > 0)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_level_previous_page, "<< PREVIOUS PAGE", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		else
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_random, "RANDOM LEVEL", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		for(i = start_entry; i < max; i++)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_level, instance->level_list->entry[instance->level_list->unlocked_index[i]].name, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		if(instance->level_page < (instance->level_list->unlocked_count - 1) / 9)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_level_next_page, "NEXT PAGE >>", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			pos += 32;
		}
		pos += 8;
		if(instance->game.mode == WL_GAME_MODE_MUSIC)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_level_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		}
		else
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		}
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
}

int wl_menu_proc_play_mode_extra_stories_select(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.player[0].story = i;
	instance->game.mode = WL_GAME_MODE_STORY + instance->game.player[0].story;
	instance->game.player[0].character = wl_get_character_by_name(instance->level_list->story[instance->game.player[0].story].character, data);
	if(instance->game.player[0].character >= 0)
	{
		instance->level_page = 0;
		wl_create_play_story_select_menu(data, instance->game.player[0].story);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	return 1;
}

static int _wl_get_story_completion_percentage(int i, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	int p = instance->profile->progress[i] / WL_LEVEL_UP;
	int total = instance->level_list->story[i].end_level - instance->level_list->story[i].start_level + 1;

	return ((float)(p) / (float)(total)) * 100.0 + 0.5;
}

void wl_create_play_extra_stories_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, max;
	int pos = 0;
	int start_entry;
	int percent;
	char buf[64];

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{
		start_entry = instance->level_page * 9;
		max = start_entry + 9;
		if(max > instance->level_list->unlocked_count)
		{
			max = instance->level_list->unlocked_count;
		}

		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "SELECT STORY", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC);
		pos += 48;
		for(i = 1; i < instance->level_list->story_count; i++)
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_extra_stories_select, instance->level_list->story[i].name, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			percent = _wl_get_story_completion_percentage(i, data);
			sprintf(buf, "Progress: %d%%", percent);
			t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], buf);
			pos += 32;
		}
		pos += 8;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
}

int wl_menu_proc_play_select_playlist_story(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int j;
	int story = i - 1;

	/* create playlist based on selected story */
	wl_clear_playlist(instance->playlist);
	for(j = instance->level_list->story[story].start_level; j <= instance->level_list->story[story].end_level; j++)
	{
		wl_add_level_to_playlist(instance->playlist, j);
	}
	instance->playlist_pos = 0;

	/* start music mode */
	instance->game.player[0].start_level = instance->playlist->entry[0];
	wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 60);
	wl_finish_current_song(instance->music_queue, 1.0 / 60.0);
	instance->state = WL_STATE_TO_GAME;
	return 1;
}

int wl_menu_proc_play_select_playlist_all(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int j;

	/* create playlist based on selected story */
	wl_clear_playlist(instance->playlist);
	for(j = 0; j <= instance->level_list->entry_count; j++)
	{
		wl_add_level_to_playlist(instance->playlist, j);
	}
	instance->playlist_pos = 0;

	/* start music mode */
	wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 60);
	wl_finish_current_song(instance->music_queue, 1.0 / 60.0);
	instance->state = WL_STATE_TO_GAME;

	return 1;
}

int wl_menu_proc_play_select_playlist_select(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->level_page = 0;
	wl_clear_playlist(instance->playlist);
	wl_create_play_select_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

int wl_menu_proc_play_select_playlist_shuffle(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int j;

	/* create playlist based on selected story */
	wl_clear_playlist(instance->playlist);
	for(j = 0; j < instance->level_list->unlocked_count; j++)
	{
		wl_add_level_to_playlist(instance->playlist, instance->level_list->unlocked_index[j]);
	}
	wl_shuffle_playlist(instance->playlist);
	instance->playlist_pos = 0;

	/* start music mode */
	instance->game.player[0].start_level = instance->playlist->entry[0];
	wl_fade_to(instance->effects_data, al_map_rgba_f(0.0, 0.0, 0.0, 0.0), al_map_rgba_f(0.0, 0.0, 0.0, 1.0), 60);
	wl_finish_current_song(instance->music_queue, 1.0 / 60.0);
	instance->state = WL_STATE_TO_GAME;

	return 1;
}

void wl_create_play_select_music_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i, max;
	int pos = 0;
	int start_entry;

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{
		start_entry = instance->level_page * 9;
		max = start_entry + 9;
		if(max > instance->level_list->unlocked_count)
		{
			max = instance->level_list->unlocked_count;
		}

		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "SELECT PLAYLIST", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC);
		pos += 48;

		/* add story playlists */
		for(i = 0; i < instance->level_list->story_count; i++)
		{
			if(instance->level_list->unlocked_story[i])
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_playlist_story, instance->level_list->story[i].name, (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
				pos += 32;
			}
			else
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
			}
		}

		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_playlist_all, "ALL LEVELS", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		pos += 32;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_playlist_select, "SELECT LEVEL", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		pos += 32;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_playlist_shuffle, "SHUFFLE", (void **)&instance->font[WL_FONT_16], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		pos += 32;
		pos += 8;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_select_cancel, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
}

/* game mode menu */
int wl_menu_proc_play_mode_story(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.player[0].story = 0;
	instance->game.mode = WL_GAME_MODE_STORY + instance->game.player[0].story;
	instance->game.player[0].character = wl_get_character_by_name(instance->level_list->story[0].character, data);
	if(instance->game.player[0].character >= 0)
	{
		instance->level_page = 0;
		wl_create_play_story_select_menu(data, 0);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	return 1;
}

int wl_menu_proc_play_mode_extra_stories(void * data, int i, void * p)
{
	wl_create_play_extra_stories_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

static int _wl_menu_proc_play_mode_other_select_level(void * data, int i, void * p)
{
	wl_create_play_select_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

static int _wl_menu_proc_play_mode_other_start_game(void * data, int i, void * p)
{
	_wl_start_game_other(data);

	return 1;
}

static int wl_menu_proc_play_mode_other_back(void * data, int i, void * p)
{
	wl_create_play_mode_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);

	return 1;
}

bool wl_create_play_mode_other_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char * mode_name = NULL;
	int pos = 0;

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(!instance->menu[WL_MENU_DYNAMIC])
	{
		return false;
	}
	switch(instance->game.mode)
	{
		case WL_GAME_MODE_CASUAL:
		{
			mode_name = "CASUAL MODE";
			break;
		}
		case WL_GAME_MODE_PUZZLE:
		{
			mode_name = "PUZZLE MODE";
			break;
		}
		case WL_GAME_MODE_GURU:
		{
			mode_name = "GURU MODE";
			break;
		}
	}
	t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, mode_name, (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
	pos += 48;
	t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_mode_other_select_level, "SELECT LEVEL", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
	pos += 48;
	t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], _wl_menu_proc_play_mode_other_start_game, "START GAME", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
	pos += 48;
	t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_other_back, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE);
	t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);

	return true;
}

int wl_menu_proc_play_mode_casual(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.mode = WL_GAME_MODE_CASUAL;
	instance->game.player[0].character = wl_get_character_by_name("WordLeap Guy", data);
	if(instance->game.player[0].character >= 0)
	{
		instance->level_page = 0;
		wl_create_play_mode_other_menu(data);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	return 1;
}

int wl_menu_proc_play_mode_puzzle(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.mode = WL_GAME_MODE_PUZZLE;
	instance->game.player[0].character = wl_get_character_by_name("WordLeap Guy", data);
	if(instance->game.player[0].character >= 0)
	{
		instance->level_page = 0;
		wl_create_play_mode_other_menu(data);
		wl_select_menu(WL_MENU_DYNAMIC, data);
	}
	return 1;
}

int wl_menu_proc_play_mode_guru(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.mode = WL_GAME_MODE_GURU;
	instance->game.player[0].character = wl_get_character_by_name("Guru", data);
	if(instance->game.player[0].character < 0)
	{
		instance->game.player[0].character = 0;
	}
	instance->level_page = 0;
	wl_create_play_mode_other_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);
	return 1;
}

int wl_menu_proc_play_mode_music(void * data, int i, void * p)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	instance->game.mode = WL_GAME_MODE_MUSIC;
	instance->level_page = 0;
	wl_create_play_select_music_menu(data);
	wl_select_menu(WL_MENU_DYNAMIC, data);
	return 1;
}

int wl_menu_proc_play_mode_back(void * data, int i, void * p)
{
	wl_select_menu(WL_MENU_GAME, data);
	return 1;
}

static bool _extra_stories_unlocked(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int i;

	for(i = 1; i < instance->level_list->story_count; i++)
	{
		if(instance->level_list->unlocked_story[i])
		{
			return true;
		}
	}
	return false;
}

void wl_create_play_mode_menu(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	int pos = 1;
	char * item_name;
	int percent;
	char buf[64];

	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_destroy_gui(instance->menu[WL_MENU_DYNAMIC]);
	}
	instance->menu[WL_MENU_DYNAMIC] = t3f_create_gui(0, 0);
	if(instance->menu[WL_MENU_DYNAMIC])
	{
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], NULL, "GAME TYPE", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 0, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_STATIC | T3F_GUI_ELEMENT_COPY);
		item_name = "STORY";
		if(instance->level_list->story[0].name)
		{
			item_name = instance->level_list->story[0].name;
		}
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_story, item_name, (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		percent = _wl_get_story_completion_percentage(0, data);
		sprintf(buf, "Story Mode (Progress: %d%%)", percent);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], buf);
		pos++;
		if(_extra_stories_unlocked(data))
		{
			t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_extra_stories, "EXTRA STORIES", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
			t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Play Additional Stories");
			pos++;
		}
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_casual, "CASUAL", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Make 10 Words");
		pos++;
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_puzzle, "PUZZLE", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Try to Clear the Board");
		pos++;
		if(!instance->demo)
		{
			if((instance->profile->flags & WL_PROFILE_FLAG_UNLOCKED) || instance->profile->progress[0] / WL_LEVEL_UP > instance->level_list->story[0].end_level - instance->level_list->story[0].start_level)
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_guru, "GURU", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
				t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Survive As Long As You Can");
				pos++;
			}
		}

		/* don't unlock music mode in demo */
		if(!instance->demo)
		{
//			if((instance->profiles->entry[instance->game.player[0].profile].awards & WL_PROFILE_AWARD_WIN) || (instance->profiles->entry[instance->game.player[0].profile].flags & WL_PROFILE_FLAG_UNLOCKED))
			{
				t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_music, "SCENE", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
				t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Enjoy the Sights and Sounds");
				pos++;
			}
		}
		t3f_add_gui_text_element(instance->menu[WL_MENU_DYNAMIC], wl_menu_proc_play_mode_back, "BACK", (void **)&instance->font[WL_FONT_24], WL_VIRTUAL_DISPLAY_WIDTH / 2, 48 * pos, al_map_rgba_f(0.0, 0.0, 0.0, 1.0), T3F_GUI_ELEMENT_CENTRE | T3F_GUI_ELEMENT_COPY);
		t3f_describe_last_gui_element(instance->menu[WL_MENU_DYNAMIC], "Back to Previous Menu");
		pos++;
		t3f_center_gui(instance->menu[WL_MENU_DYNAMIC], 0, WL_VIRTUAL_DISPLAY_HEIGHT);
	}
}
